from __future__ import annotations

import re
import uuid
from typing import TYPE_CHECKING

from flow.record.fieldtypes import posix_path

from dissect.target.exceptions import FilesystemError
from dissect.target.filesystems.nfs import NfsFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.nfs.client.nfs import Client as NfsClient
from dissect.target.helpers.nfs.client.nfs import NfsError
from dissect.target.helpers.nfs.nfs3 import FileHandle, NfsStat
from dissect.target.helpers.record import UnixUserRecord
from dissect.target.helpers.sunrpc.client import LocalPortPolicy, auth_unix
from dissect.target.helpers.utils import parse_options_string
from dissect.target.loaders.local import LocalLoader
from dissect.target.plugin import OperatingSystem, OSPlugin, arg, export, internal

if TYPE_CHECKING:
    import logging
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


log = get_logger(__name__)


# https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
ARCH_MAP = {
    0x00: "unknown",
    0x02: "sparc",
    0x03: "x86",
    0x08: "mips",
    0x14: "powerpc32",
    0x15: "powerpc64",
    0x16: "s390",  # and s390x
    0x28: "aarch32",  # armv7
    0x2A: "superh",
    0x32: "ia-64",
    0x3E: "x86_64",
    0xB7: "aarch64",  # armv8
    0xF3: "riscv64",
    0xF7: "bpf",
}


class UnixPlugin(OSPlugin):
    """UNIX plugin."""

    # Files to parse for user details
    PASSWD_FILES = ("/etc/passwd", "/etc/passwd-", "/etc/master.passwd")

    def __init__(self, target: Target):
        super().__init__(target)
        self._add_mounts()
        self._add_devices()
        self._hostname, self._domain = self._parse_hostname_string()
        self._hosts_dict = self._parse_hosts_string()
        self._os_release = self._parse_os_release()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        # If the ``/var`` and ``/etc`` folders exist on a filesystem it is treated as a Unix-like filesystem
        for fs in target.filesystems:
            if fs.exists("/var") and fs.exists("/etc"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.fs.mount("/", sysvol)
        return cls(target)

    @export(record=UnixUserRecord)
    @arg("--sessions", action="store_true", help="Parse syslog for recent user sessions")
    def users(self, sessions: bool = False) -> Iterator[UnixUserRecord]:
        """Yield unix user records from passwd files or syslog session logins.

        References:
            - https://manpages.ubuntu.com/manpages/oracular/en/man5/passwd.5.html
        """

        seen_users = set()

        # Yield users found in passwd files.
        for passwd_file in self.PASSWD_FILES:
            if (path := self.target.fs.path(passwd_file)).exists():
                for line in path.open("rt", errors="surrogateescape"):
                    line = line.strip()
                    if not line or line.startswith("#"):
                        continue

                    pwent = dict(enumerate(line.split(":")))

                    current_user = (pwent.get(0), pwent.get(5), pwent.get(6))
                    if current_user in seen_users:
                        continue
                    seen_users.add(current_user)
                    yield UnixUserRecord(
                        name=pwent.get(0),
                        passwd=pwent.get(1),
                        uid=pwent.get(2) or None,
                        gid=pwent.get(3) or None,
                        gecos=pwent.get(4),
                        home=posix_path(pwent.get(5)),
                        shell=pwent.get(6),
                        source=passwd_file,
                        _target=self.target,
                    )

        # Find users not in passwd files by parsing recent
        # syslog ldap, kerberos and x-session logins.
        # Must be enabled using the --sessions flag
        if sessions and (path := self.target.fs.path("/var/log/syslog")).exists():
            sessions = []
            cur_session = -1
            needles = {
                ("setting HOME=", "home"),
                ("setting SHELL=", "shell"),
                ("setting USER=", "name"),
            }

            for line in path.open("rt"):
                # Detect the beginning of a new session activation in the syslog
                #
                # dbus-update-activation-environment starts a new session with
                # DBUS_SESSION_BUS_ADDRESS, XDG_RUNTIME_DIR and/or DISPLAY
                # Using DBUS_SESSION_BUS_ADDRESS seems to work fine for now.
                if "setting DBUS_SESSION_BUS_ADDRESS=" in line:
                    bus = line.split("DBUS_SESSION_BUS_ADDRESS=")[1].strip()
                    cur_session += 1
                    sessions.append({"bus": bus, "name": None, "home": None, "shell": None, "uid": None})

                # Search the following lines for more information on the previousley detected session.
                for n, k in needles:
                    if n in line and "dbus-update-activation-environment" in line:
                        sessions[cur_session][k] = line.split(n)[1].strip()

            for user in sessions:
                # Only return users we didn't already find in previously parsed passwd files and past sessions.
                current_user = (user["name"], user["home"], user["shell"])
                if current_user in seen_users:
                    continue

                seen_users.add(current_user)

                yield UnixUserRecord(
                    name=user["name"],
                    home=posix_path(user["home"]),
                    shell=user["shell"],
                    source="/var/log/syslog",
                    _target=self.target,
                )

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        if (root_path := self.target.fs.path("/root")).exists():
            yield (root_path, ("uid", 0))

        if (home_path := self.target.fs.path("/home")).exists():
            yield from ((entry, None) for entry in home_path.iterdir() if entry.is_dir())

    @export(property=True)
    def architecture(self) -> str | None:
        return self._get_architecture(self.os)

    @export(property=True)
    def hostname(self) -> str | None:
        return self._hostname or self._hosts_dict.get("hostname", "localhost")

    @export(property=True)
    def domain(self) -> str | None:
        if self._domain and "localhost" not in self._domain:
            return self._domain

        # fall back to /etc/hosts file
        if "localhost" not in (domain := self._hosts_dict.get("hostname", "")):
            return domain or None

        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.UNIX.value

    def _parse_hostname_string(
        self, paths: list[tuple[str, Callable[[Path], str] | None]] | None = None
    ) -> tuple[str | None, str | None]:
        """Returns a tuple containing respectively the hostname and domain name portion of the path(s) specified.

        Args:
            paths (list): list of tuples with paths and callables to parse the path or None

        Returns:
            Tuple with hostname and domain strings.
        """
        hostname = None
        domain = None

        paths = paths or [
            ("/etc/hostname", None),
            ("/etc/HOSTNAME", None),
            ("/proc/sys/kernel/hostname", None),
            ("/etc/sysconfig/network", self._parse_rh_legacy),
            ("/etc/hosts", self._parse_etc_hosts),  # fallback if no other hostnames are found
        ]

        for path, callable in paths:
            if not (path := self.target.fs.path(path)).exists():
                continue

            hostname = callable(path) if callable else path.open("rt").read().rstrip()

            if hostname and "." in hostname:
                hostname, domain = hostname.split(".", maxsplit=1)

            break  # break whenever a valid hostname is found

        # Can be an empty string due to splitting of hostname and domain
        return hostname or None, domain or None

    def _parse_rh_legacy(self, path: Path) -> str | None:
        hostname = None
        file_contents = path.open("rt").readlines()
        for line in file_contents:
            if not line.startswith("HOSTNAME"):
                continue
            _, _, hostname = line.rstrip().partition("=")
        return hostname

    def _parse_etc_hosts(self, path: Path) -> str | None:
        for line in path.open("rt"):
            if line.startswith(("127.0.0.1 ", "::1 ")) and "localhost" not in line:
                return line.split(" ")[1]
        return None

    def _parse_hosts_string(self, paths: list[str] | None = None) -> dict[str, str]:
        paths = paths or ["/etc/hosts"]
        hosts_string = {}

        for path in paths:
            for fs in self.target.filesystems:
                if fs.exists(path):
                    for line in fs.path(path).open("rt").readlines():
                        if not (line := line.split()):
                            continue
                        if line[0].startswith(("127.0.", "::1")):
                            hosts_string = {"ip": line[0], "hostname": line[1]}
        return hosts_string

    def _add_mounts(self) -> None:
        fstab = self.target.fs.path("/etc/fstab")

        for dev_id, volume_name, mount_point, fs_type, options in parse_fstab(fstab, self.target.log):
            # Mount nfs, but only when target has been mapped by the `LocalLoader`
            if fs_type == "nfs":
                self._add_nfs(dev_id, volume_name, mount_point)
                continue

            opts = parse_options_string(options)
            subvol = opts.get("subvol", None)
            subvolid = opts.get("subvolid", None)
            for fs in self.target.filesystems:
                fs_id = None
                fs_subvol = None
                fs_subvolid = None
                fs_last_mount = None
                fs_volume_name = None
                vol_volume_name = fs.volume.name if fs.volume and not isinstance(fs.volume, list) else None

                if fs.__type__ == "xfs":
                    fs_id = fs.xfs.uuid
                    fs_volume_name = fs.xfs.name
                elif fs.__type__ == "ext":
                    fs_id = fs.extfs.uuid
                    fs_last_mount = fs.extfs.last_mount
                    fs_volume_name = fs.extfs.volume_name
                elif fs.__type__ == "btrfs":
                    fs_id = fs.btrfs.uuid
                    fs_subvol = fs.subvolume.path
                    fs_subvolid = fs.subvolume.objectid
                elif fs.__type__ == "fat":
                    fs_id = fs.fatfs.volume_id
                    # This normalizes fs_id to comply with libblkid generated UUIDs
                    # This is needed because FAT filesystems don't have a real UUID,
                    # but instead a volume_id which is not case-sensitive
                    fs_id = fs_id[:4].upper() + "-" + fs_id[4:].upper()

                if (
                    (fs_id and (fs_id == dev_id and (subvol == fs_subvol or subvolid == fs_subvolid)))
                    or (fs_last_mount and (fs_last_mount == mount_point))
                    or (fs_volume_name and (fs_volume_name == volume_name))
                    or (vol_volume_name and (vol_volume_name == volume_name))
                ):
                    self.target.log.debug("Mounting %s (%s) at %s", fs, fs.volume, mount_point)
                    self.target.fs.mount(mount_point, fs)

    @property
    def _is_nfs_enabled(self) -> bool:
        return isinstance(self.target._loader, LocalLoader) and "enable-nfs" in self.target.path_query

    def _log_nfs_mount_disabled(self, address: str, exported_dir: str, mount_point: str) -> None:
        if isinstance(self.target._loader, LocalLoader):
            log.warning(
                "NFS mount %s:%s at %s is disabled. To enable, pass --enable-nfs to the local loader. Alternatively, add a query parameter to the target query string: local?enable-nfs",  # noqa: E501
                address,
                exported_dir,
                mount_point,
            )
        else:
            log.warning(
                "NFS mount %s:%s at %s is unavailable on a non-local target",
                address,
                exported_dir,
                mount_point,
            )

    def _add_nfs(self, address: str, exported_dir: str, mount_point: str) -> None:
        if not self._is_nfs_enabled:
            self._log_nfs_mount_disabled(address, exported_dir, mount_point)
            return

        # Try all users to see if we can access the share
        def auth_setter(nfs_client: NfsClient, filehandle: FileHandle, _: list[int]) -> None:
            users = list(self.users())
            if not users:
                self.target.log.debug(
                    "No users found, trying root for mounting NFS share %s:%s at %s", address, exported_dir, mount_point
                )
                users = [UnixUserRecord(uid=0, gid=0)]

            for user in users:
                if user.uid is None or user.gid is None:
                    continue
                auth = auth_unix("machine", user.uid, user.gid, [])
                nfs_client.rebind_auth(auth)
                try:
                    self.target.log.debug("Trying to read NFS share with uid %d and gid %d", user.uid, user.gid)
                    # Use a readdir to check if we have access.
                    # RdJ: Perhaps an ACCESS call (to be implemented) is better than READDIR
                    nfs_client.readdir(filehandle)
                except NfsError as e:
                    if e.nfsstat != NfsStat.ERR_ACCES:
                        self.target.log.warning("Reading NFS share gives %s", e.nfsstat)
                        nfs_client.close()
                        raise
                else:
                    # We have access
                    return

            self.target.log.debug("No user has access to NFS share %s:%s at %s", address, exported_dir, mount_point)
            raise FilesystemError(f"No user has access to NFS share {address}:{exported_dir} at {mount_point}")

        try:
            self.target.log.debug("Mounting NFS share %s at %s", exported_dir, mount_point)
            nfs = NfsFilesystem.connect(address, exported_dir, auth_setter, LocalPortPolicy.PRIVILEGED)
            self.target.fs.mount(mount_point, nfs)
        except Exception as e:
            self.target.log.warning("Failed to mount NFS share %s:%s at %s", address, exported_dir, mount_point)
            self.target.log.debug("", exc_info=e)

    def _add_devices(self) -> None:
        """Add some virtual block devices to the target.

        Currently only adds LVM devices.
        """
        vfs = self.target.fs.append_layer()

        for volume in self.target.volumes:
            if volume.vs and volume.vs.__type__ == "lvm":
                vfs.map_file_fh(f"/dev/{volume.raw.vg.name}/{volume.raw.name}", volume)

    def _parse_os_release(self, glob: str | None = None) -> dict[str, str]:
        """Parse files containing Unix version information.

        Not all these files are equal. Generally speaking these files are
        either key=value files or contain just one line.

        Examples of key=value pair structured release files are:
        - /etc/os-release
        - /usr/lib/os-release
        - /etc/lsb-release

        Examples of sparse release files:
        - /etc/fedora-release
        - /etc/centos-release
        - /etc/redhat-release
        - /etc/SuSE-release

        Examples of bsd version files:
        - /bin/freebsd-version
        """
        glob = glob or "/etc/*-release"

        os_release = {}

        for path in self.target.fs.glob(glob):
            if self.target.fs.path(path).is_file():
                with self.target.fs.path(path).open("rt") as release_file:
                    for line in release_file:
                        if line.startswith("#"):
                            continue

                        if "=" not in line:
                            os_release["DISTRIB_DESCRIPTION"] = line.strip()

                        else:
                            try:
                                name, value = line.split("=", maxsplit=1)
                                os_release[name] = value.replace('"', "").replace("\n", "")
                            except ValueError:
                                continue
        return os_release

    def _get_architecture(self, os: str = "unix", path: Path | str = "/bin/ls") -> str | None:
        """Determine architecture by reading an ELF header of a binary on the target.

        References:
            - https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#ISA
        """

        if not isinstance(path, TargetPath):
            for fs in [self.target.fs, *self.target.filesystems]:
                if (path := fs.path(path)).exists():
                    break

        if not path.exists():
            return None

        fh = path.open("rb")
        fh.seek(4)  # ELF - e_ident[EI_CLASS]
        bits = fh.read(1)[0]

        fh.seek(18)  # ELF - e_machine
        e_machine = int.from_bytes(fh.read(2), "little")
        arch = ARCH_MAP.get(e_machine, "unknown")

        return f"{arch}_32-{os}" if bits == 1 and arch[-2:] != "32" else f"{arch}-{os}"


def parse_fstab(
    fstab: TargetPath,
    log: logging.Logger = log,
) -> Iterator[tuple[uuid.UUID | str, str, str, str, str]]:
    """Parse fstab file and return a generator that streams the details of entries,
    with unsupported FS types and block devices filtered away.
    """

    SKIP_FS_TYPES = (
        "swap",
        "tmpfs",
        "devpts",
        "sysfs",
        "procfs",
        "overlayfs",
    )

    if not fstab.exists():
        return

    for entry in fstab.open("rt"):
        entry = entry.strip()
        if entry.startswith("#"):
            continue

        entry_parts = re.split(r"\s+", entry)

        if len(entry_parts) != 6:
            continue

        dev, mount_point, fs_type, options, _, _ = entry_parts

        if fs_type in SKIP_FS_TYPES:
            log.warning("Skipped FS type: %s, %s, %s", fs_type, dev, mount_point)
            continue

        dev_id = None
        volume_name = None
        if dev.startswith(("/dev/mapper", "/dev/gpt")):
            volume_name = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/disk/by-uuid"):
            dev_id = dev.rsplit("/")[-1]
        elif dev.startswith("/dev/") and dev.count("/") == 3:
            # When composing a vg-lv name, LVM2 replaces hyphens with double hyphens in the vg and lv names
            # Emulate that here when combining the vg and lv names
            volume_name = "-".join(part.replace("-", "--") for part in dev.rsplit("/")[-2:])
        elif dev.startswith("UUID="):
            dev_id = dev.split("=")[1]
        elif dev.startswith("LABEL="):
            volume_name = dev.split("=")[1]
        elif fs_type == "nfs":
            # Put the nfs server address in dev_id and the root path in volume_name
            dev_id, sep, volume_name = dev.partition(":")
            if sep != ":":
                log.warning("Invalid NFS mount: %s", dev)
                continue
        else:
            log.warning("Unsupported mount device: %s %s", dev, mount_point)
            continue

        if mount_point == "/":
            continue

        if dev_id:
            try:
                dev_id = uuid.UUID(dev_id)
            except ValueError:
                pass

        yield dev_id, volume_name, mount_point, fs_type, options



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from dissect.hypervisor.descriptor.vmx import VMX

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, alias, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

VmwareDragAndDropRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "virtualization/vmware/clipboard",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)

VmwareVirtualMachineRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "virtualization/vmware/vm",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("boolean", "is_clone"),
        ("boolean", "is_favorite"),
        ("string", "state"),
        ("string", "uuid"),
        ("string", "annotation"),
        ("string[]", "mac_addresses"),
        ("path[]", "disks"),
        ("path[]", "sources"),
    ],
)


INVENTORY_PATHS = [
    # Windows
    "AppData/Roaming/VMware/inventory.vmls",
    # Linux
    ".vmware/inventory.vmls",
    # macOS (Fusion)
    "Library/Application Support/VMware Fusion/vmInventory",
]


DND_PATHS = [
    # Windows
    "AppData/Local/Temp/VmwareDND",
    # Linux
    ".cache/vmware/drag_and_drop",
]


class VmwareWorkstationPlugin(Plugin):
    """VMware Workstation plugin."""

    __namespace__ = "vmware"

    def __init__(self, target: Target):
        super().__init__(target)
        self.inventories = list(find_vm_inventory(self.target))
        self.dnd_dirs = list(self.find_dnd_dirs())

    def check_compatible(self) -> None:
        if not self.dnd_dirs and not self.inventories:
            raise UnsupportedPluginError("No VMware Workstation artifact(s) found")

    def find_dnd_dirs(self) -> Iterator[tuple[UserDetails, Path]]:
        for user_details in self.target.user_details.all_with_home():
            for dnd_path in DND_PATHS:
                if (dnd_dir := user_details.home_path.joinpath(dnd_path)).exists():
                    yield user_details, dnd_dir

    @alias("draganddrop")
    @export(record=VmwareDragAndDropRecord)
    def clipboard(self) -> Iterator[VmwareDragAndDropRecord]:
        """Yield cached VMware Workstation drag-and-drop file artifacts."""

        for user_details, dnd_dir in self.dnd_dirs:
            for file in dnd_dir.rglob("*/*"):
                if file.is_dir():
                    continue

                yield VmwareDragAndDropRecord(
                    ts=file.lstat().st_mtime,
                    path=file,
                    _user=user_details.user,
                    _target=self.target,
                )

    @export(record=VmwareVirtualMachineRecord)
    def config(self) -> Iterator[VmwareVirtualMachineRecord]:
        """Yield VMware Workstation Virtual Machine inventory configurations.

        Parses ``inventory.vmls`` and ``.vmx`` descriptor files. Does not parse newer ``.vmxf`` XML files.
        Does not support older ``vmAutoStart.xml`` or ``vmInventory.xml`` formats.

        References:
            - https://sanbarrow.com/vmx/vmx-network-advanced.html
        """

        for inventory, user_details in self.inventories:
            for config in parse_inventory_file(inventory).values():
                vmx_config = {}

                if (vmx_path := self.target.fs.path(config.get("config"))).exists():
                    vmx = VMX.parse(vmx_path.read_text())
                    vmx_config = vmx.attr

                yield VmwareVirtualMachineRecord(
                    ts=vmx_path.lstat().st_mtime if vmx_config else inventory.lstat().st_mtime,
                    # Inventory config fields
                    name=config.get("DisplayName"),
                    is_clone=config.get("IsClone", "").lower() == "true",
                    is_favorite=config.get("IsFavorite", "").lower() == "true",
                    state=config.get("State"),
                    uuid=config.get("UUID").replace(" ", ""),
                    # VMX config fields
                    annotation=vmx_config.get("annotation"),
                    mac_addresses={v for k, v in vmx_config.items() if k.endswith(("generatedAddress", "address"))},
                    disks=vmx.disks(),
                    # Metadata
                    sources=[inventory, vmx_path],
                    _user=user_details.user if user_details else None,
                    _target=self.target,
                )


def find_vm_inventory(target: Target) -> Iterator[tuple[Path, UserDetails]]:
    """Search for ``inventory.vmls`` files in user home folders."""

    for user_details in target.user_details.all_with_home():
        for inv_path in INVENTORY_PATHS:
            if (inv_file := user_details.home_path.joinpath(inv_path)).exists():
                yield inv_file, user_details


def parse_inventory_file(inventory: Path) -> dict[str, Any] | None:
    """Parse a single ``inventory.vmls`` (Windows, Linux) or ``vmInventory`` (macOS) file."""

    config = defaultdict(dict)
    with inventory.open("rt") as fh:
        for line in map(str.strip, fh):
            if not line or line.startswith("."):
                continue
            full_key, value = map(str.strip, line.split("=", 1))
            vm, key = full_key.split(".", 1)

            # Only process vmlist entries, not index entries
            if "vmlist" not in vm:
                continue

            config[vm][key] = value.strip('"')
        return config



from __future__ import annotations

import random
import socket
import sys
from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Generic, TypeVar

from dissect.target.helpers.sunrpc import sunrpc
from dissect.target.helpers.sunrpc.serializer import (
    AuthFlavor,
    AuthNullSerializer,
    AuthSerializer,
    AuthUnixSerializer,
    Deserializer,
    MessageSerializer,
    PortMappingSerializer,
    Serializer,
    UInt32Serializer,
)

if TYPE_CHECKING:
    from types import TracebackType

    from typing_extensions import Self

    from dissect.target.helpers.nfs.nfs3 import ProcedureDescriptor


Credentials = TypeVar("Credentials")
Verifier = TypeVar("Verifier")
ConCredentials = TypeVar("ConCredentials")
ConVerifier = TypeVar("ConVerifier")
Params = TypeVar("Params")
Results = TypeVar("Results")


@dataclass
class AuthScheme(Generic[Credentials, Verifier]):
    flavor: int
    credentials: Credentials
    verifier: Verifier
    credentials_serializer: AuthSerializer[Credentials]
    verifier_serializer: AuthSerializer[Verifier]


def auth_null() -> AuthScheme[sunrpc.AuthNull, sunrpc.AuthNull]:
    return AuthScheme(
        AuthFlavor.AUTH_NULL.value, sunrpc.AuthNull(), sunrpc.AuthNull(), AuthNullSerializer(), AuthNullSerializer()
    )


def auth_unix(machine: str | None, uid: int, gid: int, gids: list[int]) -> AuthScheme[sunrpc.AuthUnix, sunrpc.AuthNull]:
    stamp = random.randint(0, 2**32 - 1)

    if machine is None:
        machine = "dissect"

    return AuthScheme(
        AuthFlavor.AUTH_UNIX.value,
        sunrpc.AuthUnix(stamp, machine, uid, gid, gids),
        sunrpc.AuthNull(),
        AuthUnixSerializer(),
        AuthNullSerializer(),
    )


class MismatchXidError(Exception):
    pass


class UnexpectedResponse(Exception):
    pass


class IncompleteMessage(Exception):
    pass


class InvalidPortMapping(Exception):
    pass


class AbstractClient(ABC):
    @abstractmethod
    def call(
        self,
        proc_desc: ProcedureDescriptor,
        params: Params,
        params_serializer: Serializer[Params],
        result_deserializer: Deserializer[Results],
    ) -> Results:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def rebind_auth(self, auth: AuthScheme[ConCredentials, ConVerifier]) -> None:
        pass


class LocalPortPolicy(IntEnum):
    """Policy for binding to a local port."""

    ANY = 0  # Bind to any free port
    PRIVILEGED = -1  # Bind to the first free privileged port


class Client(AbstractContextManager, AbstractClient, Generic[Credentials, Verifier]):
    PMAP_PORT = 111
    TCP_KEEPIDLE = 60
    TCP_KEEPINTVL = 10
    TCP_KEEPCNT = 3

    @classmethod
    def connect_port_mapper(
        cls, hostname: str, timeout_in_seconds: float | None = 5.0
    ) -> Client[sunrpc.AuthNull, sunrpc.AuthNull]:
        """Connect to the port mapper service on a remote host."""
        return cls.connect(
            hostname,
            cls.PMAP_PORT,
            auth_null(),
            local_port=0,
            timeout_in_seconds=timeout_in_seconds,
        )

    @classmethod
    def connect(
        cls,
        hostname: str,
        port: int,
        auth: AuthScheme[ConCredentials, ConVerifier],
        local_port: int | LocalPortPolicy = 0,
        timeout_in_seconds: float | None = 5.0,
    ) -> Client[ConCredentials, ConVerifier]:
        """Connect to a RPC server.

        Args:
            hostname: The remote hostname.
            port: The remote port.
            auth: The authentication scheme.
            local_port: The local port to bind to.
                If equal to ``LocalPortPolicy.PRIVILEGED`` or -1, bind to the first free privileged port.
                If equal to ``LocalPortPolicy.ANY`` or 0, bind to any free port.
                Otherwise, bind to the specified port.
            timeout_in_seconds: The timeout for making the connection.
        """

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if not sys.platform.startswith("darwin"):
            # MacOS does not support TCP_KEEPIDLE, and TCP_KEEPALIVE is only available in Python 3.10
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, cls.TCP_KEEPIDLE)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, cls.TCP_KEEPINTVL)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, cls.TCP_KEEPCNT)

        local_port_int = local_port.value if isinstance(local_port, LocalPortPolicy) else local_port
        if local_port_int == LocalPortPolicy.PRIVILEGED:
            cls._bind_free_privileged_port(sock)
        else:
            sock.bind(("", local_port_int))

        sock.settimeout(timeout_in_seconds)
        sock.connect((hostname, port))
        sock.settimeout(None)  # Back to blocking mode

        return Client(sock, auth)

    @classmethod
    def _bind_free_privileged_port(cls, sock: socket.socket) -> None:
        """Bind to a free privileged port (1-1023)."""
        for port in range(1, 1024):
            try:
                return sock.bind(("", port))
            except OSError:  # noqa: PERF203
                continue

        raise OSError("No free privileged port available")

    def __init__(self, sock: socket.socket, auth: AuthScheme[Credentials, Verifier], fragment_size: int = 8192):
        self._sock = sock
        self._auth = auth
        self._fragment_size = fragment_size
        self._xid = 1

    def __enter__(self) -> Self:
        """Return ``self`` upon entering the runtime context."""
        return self  # type: Necessary for type checker

    def __exit__(self, _: type[BaseException] | None, __: BaseException | None, ___: TracebackType | None) -> bool:
        self.close()
        return False  # Reraise exceptions

    def close(self) -> None:
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass  # Ignore errors if the socket is already closed
        self._sock.close()

    def rebind_auth(self, auth: AuthScheme[ConCredentials, ConVerifier]) -> Client[ConCredentials, ConVerifier]:
        """Return a new client with the same socket, but different authentication credentials."""
        fd = self._sock.detach()  # Move underlying file descriptor to new socket
        new_sock = socket.socket(fileno=fd)
        return Client(new_sock, auth, self._fragment_size)

    def query_port_mapping(self, program: int, version: int) -> int:
        """Query port number of specified program and version."""
        arg = sunrpc.PortMapping(program=program, version=version, protocol=sunrpc.Protocol.TCP)
        result = self.call(sunrpc.GetPortProc, arg, PortMappingSerializer(), UInt32Serializer())
        if result == 0:
            raise InvalidPortMapping("Invalid port mapping")
        return result

    def call(
        self,
        proc_desc: ProcedureDescriptor,
        params: Params,
        params_serializer: Serializer[Params],
        result_deserializer: Deserializer[Results],
    ) -> Results:
        """Synchronously call an RPC procedure and return the result."""

        call_body = sunrpc.CallBody(
            proc_desc.program,
            proc_desc.version,
            proc_desc.procedure,
            self._auth.credentials,
            self._auth.verifier,
            params,
        )
        message = sunrpc.Message(self._xid, call_body)
        message_serializer = MessageSerializer(
            params_serializer, result_deserializer, self._auth.credentials_serializer, self._auth.verifier_serializer
        )
        request_payload = message_serializer.serialize(message)
        self._send(request_payload)

        response_payload = self._receive()
        response = message_serializer.deserialize_from_bytes(response_payload)
        if response.xid != self._xid:
            raise MismatchXidError("Invalid response xid")
        if not isinstance(response.body, sunrpc.AcceptedReply):
            raise UnexpectedResponse("Unexpected response type")
        if response.body.stat != sunrpc.AcceptStat.SUCCESS:
            raise UnexpectedResponse("Call failed")

        self._xid += 1
        return response.body.results

    def _send(self, data: bytes) -> None:
        data_size = len(data)
        offset = 0

        # Messages are split into fragments: 4 bytes for the fragment size, followed by the fragment data
        while offset < data_size:
            fragment = data[offset : offset + self._fragment_size]
            fragment_size = len(fragment)

            fragment_header = fragment_size
            if offset + fragment_size == data_size:
                fragment_header = fragment_header | 0x80000000  # MSB set to indicate last fragment

            chunk = fragment_header.to_bytes(4, "big") + fragment
            self._sock.sendall(chunk)
            offset += fragment_size

    def _receive(self) -> bytes:
        fragments = []

        while True:
            header = self._sock.recv(4)
            if not header:
                raise IncompleteMessage("Expected a fragment header")

            fragment_header = int.from_bytes(header, "big")
            fragment_size = fragment_header & 0x7FFFFFFF
            while fragment_size > 0:
                fragment = self._sock.recv(fragment_size)
                fragments.append(fragment)
                fragment_size -= len(fragment)

            # Check for last fragment
            if fragment_header & 0x80000000:
                return b"".join(fragments)

    def __del__(self):
        self._sock.close()



from __future__ import annotations

import pathlib
import tempfile
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.windows.log.intunemanagementextension import (
    IntuneManagementExtensionLogParserPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

LOG_DIR = "ProgramData/Microsoft/IntuneManagementExtension/Logs"


def test_intunemanagementextension_check_compatible_missing_dir(
    target_win: Target,
) -> None:
    """Plugin should raise if the log directory is missing."""
    from dissect.target.exceptions import UnsupportedPluginError

    with pytest.raises(UnsupportedPluginError):
        target_win.add_plugin(IntuneManagementExtensionLogParserPlugin)


def test_intunemanagementextension_check_compatible_missing_logs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Plugin should raise if the directory exists but contains no log files."""
    from dissect.target.exceptions import UnsupportedPluginError

    with tempfile.TemporaryDirectory() as tmpdir:
        fs_win.map_dir(LOG_DIR, pathlib.Path(tmpdir))
        with pytest.raises(UnsupportedPluginError):
            target_win.add_plugin(IntuneManagementExtensionLogParserPlugin)


def test_intunemanagementextension_check_compatible_success(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Plugin should load when an IntuneManagementExtension log exists."""
    fs_win.map_file(
        f"{LOG_DIR}/IntuneManagementExtension.log",
        absolute_path("_data/plugins/os/windows/log/intunemanagementextension/IntuneManagementExtension.log"),
    )

    target_win.add_plugin(IntuneManagementExtensionLogParserPlugin)


def test_intunemanagementextension_parsing_main_and_rotated(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Should parse valid IntuneManagementExtension log entries."""

    base = "_data/plugins/os/windows/log/intunemanagementextension"

    fs_win.map_file(
        f"{LOG_DIR}/IntuneManagementExtension.log",
        absolute_path(f"{base}/IntuneManagementExtension.log"),
    )
    fs_win.map_file(
        f"{LOG_DIR}/IntuneManagementExtension-20230101-100000.log",
        absolute_path(f"{base}/IntuneManagementExtension-20230101-100000.log"),
    )

    target_win.add_plugin(IntuneManagementExtensionLogParserPlugin)
    records = list(target_win.intunemanagementextension())

    assert len(records) == 3

    first, second, third = records

    expected_type = "windows_intune_managementextension_log"
    assert type(first).__name__ == expected_type
    assert type(second).__name__ == expected_type
    assert type(third).__name__ == expected_type

    assert first.component == "IntuneManagementExtension"
    assert second.component == "IntuneManagementExtension"
    assert third.component == "IntuneManagementExtension"

    assert first.thread == "1002"
    assert first.type == "2"
    assert first.context == "Agent"
    assert "Another entry." in first.message
    assert first.file_origin.endswith("IntuneManagementExtension.log")

    assert second.thread == "1003"
    assert second.type == "3"
    assert second.context == "Policy"
    assert "YYYY-MM-DD format test." in second.message
    assert second.file_origin.endswith("IntuneManagementExtension.log")

    assert third.thread == "1001"
    assert third.type == "1"
    assert third.context == "ManagedSoftware"
    assert "Test message one." in third.message
    assert third.file_origin.endswith("IntuneManagementExtension-20230101-100000.log")



from __future__ import annotations

import base64
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import path

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.apps.ssh.putty import PuTTYPlugin, construct_public_key
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_putty_plugin_ssh_host_keys_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    ssh_host_keys = absolute_path("_data/plugins/apps/ssh/putty/sshhostkeys")
    fs_unix.map_file("/root/.putty/sshhostkeys", ssh_host_keys)

    target_unix_users.add_plugin(PuTTYPlugin)
    records = list(target_unix_users.putty.known_hosts())

    assert len(records) == 3

    assert records[0].mtime_ts is not None
    assert records[0].host == "192.168.123.130"
    assert records[0].port == 22

    assert records[0].key_type == "ssh-ed25519"
    assert records[0].public_key_pem == base64.b64decode(
        "AAAAC3NzaC1lZDI1NTE5AAAAIHUl23i/4p/7xcZnNPDK+Dr+A539zpEEXutrm/tESFYq"
    )
    assert records[0].fingerprint.md5 == "16da68cfc0e1b7954c84147a385be2b6"
    assert records[0].fingerprint.sha1 == "11e647366ef9b74feb55da7f8507a7180123eed0"
    assert records[0].fingerprint.sha256 == "0ecaca9db7d166fd55f84ad775074bd6c743ae2d26e6dac10e60e88efbfcc01d"

    assert records[0].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[0].username == "root"

    assert records[1].mtime_ts is not None
    assert records[1].host == "example.com"
    assert records[1].port == 22

    assert records[1].key_type == "ecdsa-sha2-nistp256"
    assert records[1].public_key_pem == base64.b64decode(
        "AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBNhJOmUcnvcVIizc67M3bS0/2Gz43YPHYCRaogqAgHvgVUartyfsl5nZBoFR3FiUc9kTJJUPybWbqXaGLEthjfM="
    )
    assert records[1].fingerprint.md5 == "e23736ef80f12b672cf217c390a1acfb"
    assert records[1].fingerprint.sha1 == "93a03a809a02091aab194b04c836c37298c0602a"
    assert records[1].fingerprint.sha256 == "53d7097c310ce329d554dfb533c81f81d955facae8f46e9d7cfa4db58304c674"

    assert records[1].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[1].username == "root"

    assert records[2].mtime_ts is not None
    assert records[2].host == "1.2.3.4"
    assert records[2].port == 1234

    assert records[2].key_type == "rsa2"
    assert records[2].public_key_pem == base64.b64decode(
        "AAAAB3NzaC1yc2EAAAADAQABAAABgQCmYvCKueiA3lyhZuW0OAObOZvf0g957H0wGhGi3BruZSDpa9UQhsVNtgPZfuyx/yYWUIhQ4VSz+qw00IhZJt9myRNrXCxIABM/qubTghrdjRU+ydb0J9uTBIRz+ys/0dr0dg2Gc7C5w9+E3EpRit6VCZfTi8mNYZi+/GC12VYIDsqY9/4D4xloHJP+fs1mCNnY12VtJuIWw281fKaTCxm3H95NOBbAE6sAwr8H5lMLU34D4DcA5ZIG5F48yDmYUkILtN4qLAZMl4ZlQ6KpcxMYCN6DPj2NwR8JHzr+mkwjmGqbs8tjYD/KRYL2sYhK6Fdx7wvw5djQeVTEnhjxoeiCYNFAGkzB9BfeP4N0K2rp6F3NktFU4WIc6nLsX5G1LEWuAvAxwg+v6y4YzmpXHP+WrRHKyhS+B64aLpmD7AAJzPwSyqtt5+8SY3z7EMCpYBGz8uhdYzaY0U+qWFHtIfI5GLrl2akYBzH9t/YPAF6wdTuSh4jxVe2IVrd1x7g23Zs="
    )
    assert records[2].fingerprint.md5 == "539539a9dec6afa0c8d9cdbb9357ed5e"
    assert records[2].fingerprint.sha1 == "c0f0c0a51ed7ae4c2b271f0426554f837f9508d0"
    assert records[2].fingerprint.sha256 == "c2d4bd813cdf1eec38ea8c7bbf0c56b129f9db3e78627d5b7ebfae8db60c5eb6"

    assert records[2].path == path.from_posix("/root/.putty/sshhostkeys")
    assert records[2].username == "root"


def test_putty_plugin_ssh_host_keys_windows(
    target_win_users: Target, fs_win: VirtualFilesystem, hive_hku: VirtualHive
) -> None:
    key_name = "Software\\SimonTatham\\PuTTY\\SshHostKeys"
    key = VirtualKey(hive_hku, key_name)
    key.add_value(
        "ssh-ed25519@22:192.168.123.130",
        VirtualValue(
            hive_hku,
            "ssh-ed25519@22:192.168.123.130",
            "0x12812d95024c0d8683879fd38f977e9caa7e733334f75965bc50eba6e872e70a,0x2a564844fb9b6beb5e0491cefd9d03fe3af8caf03467c6c5fb9fe2bf78db2575",
        ),
    )
    hive_hku.map_key(key_name, key)
    target_win_users.add_plugin(PuTTYPlugin)

    records = list(target_win_users.putty.known_hosts())

    assert len(records) == 1

    assert records[0].host == "192.168.123.130"
    assert records[0].port == 22
    assert records[0].key_type == "ssh-ed25519"
    assert records[0].public_key_pem == base64.b64decode(
        "AAAAC3NzaC1lZDI1NTE5AAAAIHUl23i/4p/7xcZnNPDK+Dr+A539zpEEXutrm/tESFYq"
    )
    assert records[0].path == path.from_windows("Software\\SimonTatham\\PuTTY\\SshHostKeys")


def test_putty_plugin_saved_sessions_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    sessions_folder = absolute_path("_data/plugins/apps/ssh/putty/sessions")
    fs_unix.map_dir("/root/.putty/sessions", sessions_folder)

    target_unix_users.add_plugin(PuTTYPlugin)
    records = list(target_unix_users.putty.sessions())

    assert len(records) == 1

    assert records[0].ts == datetime.fromtimestamp(
        absolute_path("_data/plugins/apps/ssh/putty/sessions/example-saved-session").stat().st_mtime,
        tz=timezone.utc,
    )
    assert records[0].session_name == "example-saved-session"
    assert records[0].protocol == "ssh"
    assert records[0].host == "192.168.123.130"
    assert records[0].user == "saveduser"
    assert records[0].port == 22
    assert records[0].remote_command == ""
    assert records[0].port_forward == ""
    assert records[0].manual_ssh_host_keys == ""
    assert records[0].path == path.from_posix("/root/.putty/sessions/example-saved-session")
    assert records[0].username == "root"


def test_putty_plugin_saved_sessions_windows(
    target_win_users: Target, fs_win: VirtualFilesystem, hive_hku: VirtualHive
) -> None:
    key_name = "Software\\SimonTatham\\PuTTY\\Sessions\\example-saved-session"
    key = VirtualKey(hive_hku, key_name)
    key.add_value("HostName", VirtualValue(hive_hku, "HostName", "example.com"))
    key.add_value("UserName", VirtualValue(hive_hku, "UserName", "user"))
    key.add_value("Protocol", VirtualValue(hive_hku, "Protocol", "ssh"))
    key.add_value("PortNumber", VirtualValue(hive_hku, "PortNumber", 22))
    key.add_value("RemoteCommand", VirtualValue(hive_hku, "RemoteCommand", ""))
    key.add_value("PortForwardings", VirtualValue(hive_hku, "PortForwardings", ""))
    key.add_value("SSHManualHostKeys", VirtualValue(hive_hku, "SSHManualHostKeys", ""))
    hive_hku.map_key(key_name, key)
    target_win_users.add_plugin(PuTTYPlugin)

    records = list(target_win_users.putty.sessions())

    assert len(records) == 1

    assert records[0].session_name == "example-saved-session"
    assert records[0].protocol == "ssh"
    assert records[0].host == "example.com"
    assert records[0].user == "user"
    assert records[0].port == 22
    assert records[0].remote_command == ""
    assert records[0].port_forward == ""
    assert records[0].manual_ssh_host_keys == ""
    assert records[0].path == path.from_windows("Software\\SimonTatham\\PuTTY\\Sessions\\example-saved-session")


def test_construct_public_key() -> None:
    with pytest.raises(TypeError, match="Invalid key_type or iv"):
        construct_public_key(None, None)

    assert construct_public_key("unsupported", "some-iv") == (
        "some-iv",
        (None, None, None),
    )



from __future__ import annotations

import stat
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence import ad1

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil, keychain

if TYPE_CHECKING:
    from collections.abc import Iterator


class AD1Filesystem(Filesystem):
    __type__ = "ad1"

    def __init__(self, fh: BinaryIO | list[BinaryIO] | Path | list[Path], *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.ad1 = ad1.AD1(fh)

        if self.ad1.is_adcrypt():
            keys = keychain.get_keys_for_provider(self.__type__) + keychain.get_keys_without_provider()

            if not keys:
                raise ValueError("Failed to unlock ADCRYPT: no key(s) provided")

            for key in keys:
                try:
                    if key.key_type == keychain.KeyType.PASSPHRASE:
                        self.ad1.unlock(passphrase=key.value)
                    elif key.key_type == keychain.KeyType.FILE and (path := Path(key.value)).is_file():
                        self.ad1.unlock(private_key=path)
                except ValueError:  # noqa: PERF203
                    pass

            if self.ad1.is_locked():
                raise ValueError("Failed to unlock ADCRYPT using provided key(s)")

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return fh.read(16) == b"ADSEGMENTEDFILE\x00"

    def get(self, path: str) -> FilesystemEntry:
        return AD1FilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(self, path: str, entry: ad1.FileEntry | None = None) -> ad1.FileEntry:
        try:
            return self.ad1.entry(path, entry)
        except ad1.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except ad1.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except ad1.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except ad1.Error as e:
            raise FileNotFoundError(path) from e


class AD1DirEntry(DirEntry):
    fs: AD1Filesystem
    entry: ad1.FileEntry

    def get(self) -> AD1FilesystemEntry:
        return AD1FilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class AD1FilesystemEntry(FilesystemEntry):
    fs: AD1Filesystem
    entry: ad1.FileEntry

    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        entry = self.fs._get_entry(path, self.entry)
        return AD1FilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self.entry.open()

    def scandir(self) -> Iterator[AD1DirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self.entry.iterdir():
            yield AD1DirEntry(self.fs, self.path, entry.name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(self.path)
        return self.entry.readlink()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        if self.is_symlink():
            mode = stat.S_IFLNK
        elif self.is_file():
            mode = stat.S_IFREG
        else:
            mode = stat.S_IFDIR

        st_info = fsutil.stat_result(
            [
                mode | 0o777,
                self.entry.offset,
                id(self.fs),
                1,  # nlink
                0,  # uid
                0,  # gid
                self.entry.size,
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        st_info.st_birthtime = self.entry.btime.timestamp()
        return st_info



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING
from uuid import UUID

from dissect.cstruct import cstruct
from dissect.database.sqlite3 import SQLite3
from dissect.util.ts import wintimestamp
from flow.record import GroupedRecord

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import WindowsUserRecord, create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

appdb_def = """
typedef struct {
    char     Magic[4];                 // Always b"DNPW"
    DWORD    Version;                  // Versions 1 (win8) and 3 (win10) seem to exist
    QWORD    Timestamp;                // According to some sources this is LastNotificationDate,
                                       // but that seems incorrect.
    DWORD    NextNotificationId;
    DWORD    Unknown;
    char     Padding[8];
} ChunkHeader;                         // size: 0x20

typedef struct {
    WORD     InUse;                    // ??
    WORD     NextTileWriteIndex;       // ??
    WORD     NextToastWriteIndex;      // ??
    BYTE     Flags[2];
} ChunkInfo;

typedef struct {
    QWORD    Timestamp1;               // ??
    QWORD    Timestamp2;               // Is this time to refresh?
    char     Uri[1024];                // Is this the correct size?
    char     Padding[0x818 - 0x410];
} PushDescriptor;                      // size: 0x818

typedef struct {
    DWORD    Id;
    DWORD    Zero;                     // ??
    QWORD    Timestamp;                // ??
    WORD     Unknown;
    WORD     DataLength;
    char     Data[DataLength];
    char     Padding[0x118 - 0x14 - DataLength];
} BadgeContent;                        // size: 0x118

typedef struct {
    DWORD    UniqueId;                 // ??
    DWORD    Zero;
    QWORD    ExpiryTime;               // The time this tile expires
    QWORD    ArrivalTime;              // The time this tile was set
    BYTE     Type;                     // ??
    BYTE     Index;
    WORD     ContentLength;
    wchar_t  Name[18];
} TileDescriptor;                      // size: 0x40

typedef struct {
    DWORD    UniqueId;                 // ??
    DWORD    Zero;
    QWORD    ExpiryTime;               // The time this toast expires
    QWORD    ArrivalTime;              // The time this toast was set
    BYTE     Type;                     // ??
    BYTE     Index;
    WORD     ContentLength;
    wchar_t  Name1[17];
    wchar_t  Name2[17];
} ToastDescriptor;                     // size: 0x60

typedef struct {
  char Content[0x1400];
} DataXML;                             // size: 0x1400

typedef struct {
    ChunkHeader     Header;            // Only populated for first chunk, else zeroed
    ChunkInfo       Info;
    PushDescriptor  Push;
    BadgeContent    Badge;
    TileDescriptor  Tiles[5];          // start @ 0x958
    DataXML         TileXml[5];

    // For the in use chunks, 0x14 ToastDiscriptors have an Index, but there
    // is space for more. Maybe this is used in case of deleted entries?
    ToastDescriptor Toasts[0x14];      // start @ 0x6e98
    char            Padding1[0x1e00];  // start @ 0x7618
    DataXML         ToastXml[0x14];    // start @ 0x9418
    char            Padding2[0x13f8];  // start @ 0x22418
} Chunk;                               // size: 0x23810
"""

c_appdb = cstruct(endian="<").load(appdb_def)

APPDB_MAGIC = b"DNPW"
NUM_APPDB_CHUNKS = 256

AppDBRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb",
    [
        ("datetime", "timestamp"),
        ("varint", "chunk_version"),
        ("varint", "next_notification_id"),
    ],
)

AppDBPushRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/push",
    [
        ("datetime", "push_ts1"),
        ("datetime", "push_ts2"),
        ("varint", "chunk_num"),
        ("uri", "push_uri"),
    ],
)

AppDBBadgeRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/badge",
    [
        ("datetime", "badge_ts"),
        ("varint", "badge_id"),
        ("varint", "chunk_num"),
        ("string", "badge_data"),
    ],
)

AppDBTileRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/tile",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "id"),
        ("varint", "chunk_num"),
        ("varint", "tile_type"),
        ("varint", "index"),
        ("string", "name"),
        ("string", "content"),
    ],
)

AppDBToastRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/appdb/toast",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "id"),
        ("varint", "chunk_num"),
        ("varint", "toast_type"),
        ("varint", "index"),
        ("string", "name1"),
        ("string", "name2"),
        ("string", "content"),
    ],
)

WpnDatabaseNotificationRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/wpndatabase",
    [
        ("datetime", "arrival_time"),
        ("datetime", "expiry_time"),
        ("varint", "order"),
        ("varint", "id"),
        ("varint", "handler_id"),
        ("string", "activity_id"),
        ("string", "type"),
        ("bytes", "payload_data"),
        ("string", "payload_type"),
        ("string", "tag"),
        ("string", "group"),
        ("varint", "boot_id"),
        ("boolean", "expires_on_reboot"),
    ],
)

WpnDatabaseNotificationHandlerRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/notification/wpndatabase/handler",
    [
        ("datetime", "created_time"),
        ("datetime", "modified_time"),
        ("varint", "id"),
        ("string", "primary_id"),
        ("string", "wns_id"),
        ("string", "handler_type"),
        ("varint", "wnf_event_name"),
        ("bytes", "system_data_property_set"),
    ],
)

NOTIFICATIONS_DIR = "AppData/Local/Microsoft/Windows/Notifications"


class NotificationsPlugin(Plugin):
    """Plugin that parses the notification databases on Windows 10 machines."""

    __namespace__ = "notifications"

    def __init__(self, target: Target):
        super().__init__(target)
        self.wpndb_files = []
        self.appdb_files = []

        for user_details in target.user_details.all_with_home():
            notification_dir = user_details.home_path.joinpath(NOTIFICATIONS_DIR)
            wpndb_file = notification_dir.joinpath("wpndatabase.db")
            appdb_file = notification_dir.joinpath("appdb.dat")

            if wpndb_file.exists():
                self.wpndb_files.append((user_details.user, wpndb_file))

            if appdb_file.exists():
                with appdb_file.open(mode="rb") as fp:
                    chunk = c_appdb.Chunk(fp)
                    if chunk.Header.Magic == APPDB_MAGIC:
                        version = chunk.Header.Version
                        if version == 3:
                            self.appdb_files.append((user_details.user, appdb_file))
                        else:
                            self.target.log.warning(
                                "Skipping %s: unsupported version %s.",
                                appdb_file,
                                version,
                            )
                            if version != 1:
                                self.target.log.warning(
                                    "Unknown appdb version %s in file %s, please consider providing us with a sample.",
                                    version,
                                    appdb_file,
                                )

    def check_compatible(self) -> None:
        if not self.wpndb_files and not self.appdb_files:
            raise UnsupportedPluginError("No or incompatible wpndatabase.db or appdb.dat files found")

    def _get_appdb_chunk_record(
        self,
        chunk: c_appdb.Chunk,
        user: WindowsUserRecord,
    ) -> AppDBRecord:
        chunk_timestamp = None
        if ts := chunk.Header.Timestamp:
            chunk_timestamp = wintimestamp(ts)

        return AppDBRecord(
            timestamp=chunk_timestamp,
            chunk_version=chunk.Header.Version,
            next_notification_id=chunk.Header.NextNotificationId,
        )

    def _get_appdb_push_record(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> AppDBPushRecord | None:
        badge_record = None
        push_uri = chunk.Push.Uri.split(b"\x00")[0]
        push_uri = push_uri.decode("utf-8", errors="surrogateescape")

        if push_uri:
            push_timestamp1 = None
            if ts := chunk.Push.Timestamp1:
                push_timestamp1 = wintimestamp(ts)
            push_timestamp2 = None
            if ts := chunk.Push.Timestamp2:
                push_timestamp2 = wintimestamp(ts)

            badge_record = AppDBPushRecord(
                push_ts1=push_timestamp1,
                push_ts2=push_timestamp2,
                chunk_num=chunk_num,
                push_uri=push_uri,
                _target=self.target,
                _user=user,
            )

        return badge_record

    def _get_appdb_badge_record(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> AppDBBadgeRecord | None:
        badge_record = None
        badge_id = chunk.Badge.Id

        if badge_id:
            badge_ts = None
            if ts := chunk.Badge.Timestamp:
                badge_ts = wintimestamp(ts)
            badge_data = chunk.Badge.Data.decode("utf-8", errors="surrogateescape")

            badge_record = AppDBBadgeRecord(
                badge_id=badge_id,
                badge_ts=badge_ts,
                chunk_num=chunk_num,
                badge_data=badge_data,
                _target=self.target,
                _user=user,
            )

        return badge_record

    def _get_appdb_tile_records(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> list[AppDBTileRecord]:
        tile_records = []
        num_tiles = len(chunk.Tiles)

        for tile_no in range(num_tiles):
            tile = chunk.Tiles[tile_no]

            if tile.UniqueId:
                tile_arrival_time = None
                if ts := tile.ArrivalTime:
                    tile_arrival_time = wintimestamp(ts)
                tile_expiry_time = None
                if ts := tile.ExpiryTime:
                    tile_expiry_time = wintimestamp(ts)
                name = tile.Name.strip("\x00")

                xml_size = tile.ContentLength
                tile_xml = chunk.TileXml[tile_no].Content[:xml_size]
                tile_xml = tile_xml.decode("utf-8", errors="surrogateescape")

                tile_record = AppDBTileRecord(
                    arrival_time=tile_arrival_time,
                    expiry_time=tile_expiry_time,
                    id=tile.UniqueId,
                    chunk_num=chunk_num,
                    tile_type=tile.Type,
                    index=tile.Index,
                    name=name,
                    content=tile_xml,
                    _target=self.target,
                    _user=user,
                )

                tile_records.append(tile_record)
        return tile_records

    def _get_appdb_toast_records(
        self,
        chunk: c_appdb.Chunk,
        chunk_num: int,
        user: WindowsUserRecord,
    ) -> list[AppDBToastRecord]:
        toast_records = []
        num_toasts = len(chunk.Toasts)

        for toast_no in range(num_toasts):
            toast = chunk.Toasts[toast_no]

            if toast.UniqueId:
                toast_arrival_time = None
                if ts := toast.ArrivalTime:
                    toast_arrival_time = wintimestamp(ts)
                toast_expiry_time = None
                if ts := toast.ExpiryTime:
                    toast_expiry_time = wintimestamp(ts)
                name1 = toast.Name1.strip("\x00")
                name2 = toast.Name2.strip("\x00")

                xml_size = toast.ContentLength
                toast_xml = chunk.ToastXml[toast_no].Content[:xml_size]
                toast_xml = toast_xml.decode("utf-8", errors="surrogateescape")

                toast_record = AppDBToastRecord(
                    arrival_time=toast_arrival_time,
                    expiry_time=toast_expiry_time,
                    id=toast.UniqueId,
                    chunk_num=chunk_num,
                    toast_type=toast.Type,
                    index=toast.Index,
                    name1=name1,
                    name2=name2,
                    content=toast_xml,
                    _target=self.target,
                    _user=user,
                )

                toast_records.append(toast_record)
        return toast_records

    @export(record=[AppDBRecord, AppDBPushRecord, AppDBBadgeRecord, AppDBTileRecord, AppDBToastRecord])
    def appdb(self) -> Iterator[GroupedRecord]:
        """Retrun the data from Windows appdb.dat file.

        This file contains data presentted to the user, pushed by external
        sources. The appdb.dat file was used from Windows 8 to Windows 10 pre
        anniversary version. This plugin only supports appdb.dat version 3 from
        Windows 10.

        References:
            - http://www.swiftforensics.com/2016/06/prasing-windows-10-notification-database.html
        """
        for user, appdb_file in self.appdb_files:
            with appdb_file.open(mode="rb") as fp:
                for chunk_num in range(NUM_APPDB_CHUNKS):
                    chunk = c_appdb.Chunk(fp)

                    if chunk.Info.InUse == 0:
                        continue
                    elif chunk.Info.InUse != 1:
                        self.target.log.warning(
                            "Unknown field value %s for chunk.Info.InUse, please consider providing us with a sample.",
                            chunk.Info.InUse,
                        )
                        continue

                    if chunk_num == 0:
                        yield self._get_appdb_chunk_record(chunk, user)

                    push_record = self._get_appdb_push_record(chunk, chunk_num, user)
                    if push_record:
                        yield push_record

                    badge_record = self._get_appdb_badge_record(chunk, chunk_num, user)
                    if badge_record:
                        yield badge_record

                    yield from self._get_appdb_tile_records(chunk, chunk_num, user)

                    yield from self._get_appdb_toast_records(chunk, chunk_num, user)

    @export(record=[WpnDatabaseNotificationRecord, WpnDatabaseNotificationHandlerRecord])
    def wpndatabase(self) -> Iterator[WpnDatabaseNotificationRecord | WpnDatabaseNotificationHandlerRecord]:
        """Returns Windows Notifications from wpndatabase.db (post Windows 10 Anniversary).

        References:
            - https://inc0x0.com/2018/10/windows-10-notification-database/
        """
        target_tz = self.target.datetime.tzinfo

        for user, wpndatabase in self.wpndb_files:
            with SQLite3(wpndatabase) as db:
                handlers = {}

                if table := db.table("NotificationHandler"):
                    for row in table.rows():
                        handlers[row["[RecordId]"]] = WpnDatabaseNotificationHandlerRecord(
                            created_time=datetime.datetime.strptime(row["[CreatedTime]"], "%Y-%m-%d %H:%M:%S").replace(
                                tzinfo=target_tz
                            ),
                            modified_time=datetime.datetime.strptime(
                                row["[ModifiedTime]"], "%Y-%m-%d %H:%M:%S"
                            ).replace(tzinfo=target_tz),
                            id=row["[RecordId]"],
                            primary_id=row["[PrimaryId]"],
                            wns_id=row["[WNSId]"],
                            handler_type=row["[HandlerType]"],
                            wnf_event_name=row["[WNFEventName]"],
                            system_data_property_set=row["[SystemDataPropertySet]"],
                            _target=self.target,
                            _user=user,
                        )

                if table := db.table("Notification"):
                    for row in table.rows():
                        record = WpnDatabaseNotificationRecord(
                            arrival_time=wintimestamp(row["[ArrivalTime]"]),
                            expiry_time=wintimestamp(row["[ExpiryTime]"]),
                            order=row["[Order]"],
                            id=row["[Id]"],
                            handler_id=row["[HandlerId]"],
                            activity_id=UUID(bytes=row["[ActivityId]"]),
                            type=row["[Type]"],
                            payload_data=row["[Payload]"],
                            payload_type=row["[PayloadType]"],
                            tag=row["[Tag]"],
                            group=row["[Group]"],
                            boot_id=row["[BootId]"],
                            expires_on_reboot=row["[ExpiresOnReboot]"] != "FALSE",
                            _target=self.target,
                            _user=user,
                        )
                        handler = handlers.get(row["[HandlerId]"])

                        if handler:
                            yield GroupedRecord("windows/notification/wpndatabase/grouped", [record, handler])
                        else:
                            yield record



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.log.lastlog import LastLogPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_lastlog_plugin(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/lastlog/lastlog")
    fs_linux.map_file("/var/log/lastlog", data_file)

    target_linux.add_plugin(LastLogPlugin)

    results = list(target_linux.lastlog())
    assert len(results) == 1

    assert results[0].ts == datetime(2021, 12, 8, 16, 14, 6, tzinfo=timezone.utc)
    assert results[0].uid == 1001
    assert results[0].ut_user is None
    assert results[0].ut_host == "127.0.0.1"
    assert results[0].ut_tty == "pts/0"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

from dissect.target.helpers.cache import Cache
from dissect.target.plugins.os.windows.amcache import AmcachePlugin
from dissect.target.plugins.os.windows.ual import UalPlugin

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path

    from dissect.target.target import Target


def test_cache_namespace(target_bare: Target) -> None:
    cache1 = Cache(AmcachePlugin.__call__, cls=AmcachePlugin)
    cache2 = Cache(UalPlugin.__call__, cls=UalPlugin)

    assert cache1.fname != cache2.fname
    assert cache1.fname == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.__call__"
    assert cache2.fname == "dissect.target.plugins.os.windows.ual.UalPlugin.__call__"

    target_bare._config.CACHE_DIR = "/tmp"
    assert (
        cache1.cache_path(target_bare, ("a", 1234)).stem
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.__call__.KCdhJywgMTIzNCk="
    )
    assert (
        cache2.cache_path(target_bare, ("b", 5678)).stem
        == "dissect.target.plugins.os.windows.ual.UalPlugin.__call__.KCdiJywgNTY3OCk="
    )


def test_cache_filename(target_win: Target) -> None:
    plugin1 = AmcachePlugin(target_win)
    plugin2 = UalPlugin(target_win)

    cache1 = Cache(AmcachePlugin.applications)
    cache2 = Cache(UalPlugin.client_access)
    cache3 = Cache(plugin1.applications)
    cache4 = Cache(plugin2.client_access)

    assert cache1.fname == cache3.fname == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications"
    assert cache2.fname == cache4.fname == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access"

    target_win._config.CACHE_DIR = "/tmp"
    assert (
        cache1.cache_path(target_win, ()).stem
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications.KCk="
    )
    assert (
        cache2.cache_path(target_win, ()).stem == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access.KCk="
    )
    assert (
        cache3.cache_path(target_win, ()).stem
        == "dissect.target.plugins.os.windows.amcache.AmcachePlugin.applications.KCk="
    )
    assert (
        cache4.cache_path(target_win, ()).stem == "dissect.target.plugins.os.windows.ual.UalPlugin.client_access.KCk="
    )


def test_cache_write_failure_behavior(target_bare: Target, tmp_path: Path) -> None:
    """Specifically tests the 'Write Path' (Cache Miss) which returns a CacheWriter.

    We verify that CacheWriter acts as an Iterator even when the underlying
    plugin returns None (stops immediately).
    """
    target_bare._config.CACHE_DIR = str(tmp_path)

    # 1. Mock Plugin with two modes
    class MockPlugin:
        def __init__(self, target: Target):
            self.target = target

        def success(self) -> Iterator[str]:
            yield "success_data"

        def failure(self) -> Iterator[str]:
            if True:
                return None
            yield "unreachable"

    plugin = MockPlugin(target_bare)

    # 2. Setup Cache wrapper
    # We force output="yield" to use LineReader/CacheWriter
    # (mimicking the behavior of RecordWriter logic in a simpler test)
    def create_wrapper(func: Callable[..., Iterator[str]]) -> Callable[..., Iterator[str]]:
        cache = Cache(func)

        def wrapper(*args: Any, **kwargs: Any) -> Iterator[str]:
            return cache.call(*args, **kwargs)

        wrapper.__output__ = "yield"
        cache.wrapper = wrapper
        return wrapper

    wrap_success = create_wrapper(MockPlugin.success)
    wrap_failure = create_wrapper(MockPlugin.failure)

    # --- SCENARIO A: Success Case (Write Path) ---
    # This creates a CacheWriter.
    # IF CacheWriter is not wrapped in iter(), next() crashes here.
    gen_success = wrap_success(plugin)

    assert next(gen_success) == "success_data"
    # Exhaust it to ensure file write completes
    list(gen_success)

    # --- SCENARIO B: Failure Case (Write Path) ---
    # It creates a CacheWriter that wraps an empty generator.
    gen_failure = wrap_failure(plugin)

    # CRITICAL CHECK:
    # 1. It must be an iterator (iter(obj) is obj)
    # 2. calling next() should raise StopIteration, NOT TypeError
    assert iter(gen_failure) is gen_failure

    # CacheWriter should be an empty iterable
    with pytest.raises(StopIteration):
        next(gen_failure)



from __future__ import annotations

from unittest.mock import MagicMock, patch

from dissect.target import plugin
from dissect.target.plugins.general.loaders import LoaderListPlugin
from dissect.target.plugins.general.plugins import (
    PluginListPlugin,
    _categorize_functions,
    _generate_plugin_tree_overview,
)


def test_plugin_description() -> None:
    description = list(_generate_plugin_tree_overview({PluginListPlugin}, False))
    assert description == ["plugins - Print all available plugins. (output: no output)"]


def test_plugin_description_compacting() -> None:
    module = {"hello": {"world": {PluginListPlugin}}}

    description = list(_generate_plugin_tree_overview(module, False))
    assert description == [
        "hello:",
        "  world:",
        "    plugins - Print all available plugins. (output: no output)",
    ]


def test_plugin_description_in_dict_multiple() -> None:
    module = {"hello": {"world": {"data": {PluginListPlugin}, "data2": {PluginListPlugin}}}}

    description = list(_generate_plugin_tree_overview(module, False))
    assert description == [
        "hello:",
        "  world:",
        "    data:",
        "      plugins - Print all available plugins. (output: no output)",
        "    data2:",
        "      plugins - Print all available plugins. (output: no output)",
    ]


def test_plugin_description_multiple_in_path() -> None:
    module = {"hello": {"world": {PluginListPlugin, LoaderListPlugin}}}
    description = list(_generate_plugin_tree_overview(module, False))
    assert description == [
        "hello:",
        "  world:",
        "    loaders - List the available loaders. (output: no output)",
        "    plugins - Print all available plugins. (output: no output)",
    ]


@patch("dissect.target.plugins.general.plugins.plugin.load")
@patch("dissect.target.plugins.general.plugins.plugin.functions")
def test_categorize_plugins(mocked_plugins: MagicMock, mocked_load: MagicMock) -> None:
    mocked_plugins.return_value = [
        plugin.FunctionDescriptor(
            name="data",
            namespace=None,
            path="something.data",
            exported=True,
            internal=False,
            findable=True,
            alias=False,
            output=None,
            method_name="data",
            module="other.root.something.data",
            qualname="DataClass",
        ),
    ]
    assert _categorize_functions() == {"something": {mocked_load.return_value}}

    mocked_plugins.return_value = [
        plugin.FunctionDescriptor(
            name="data",
            namespace=None,
            path="something.data",
            exported=True,
            internal=False,
            findable=True,
            alias=False,
            output=None,
            method_name="data",
            module="other.root.something.data",
            qualname="DataClass1",
        ),
        plugin.FunctionDescriptor(
            name="data",
            namespace=None,
            path="something.data",
            exported=True,
            internal=False,
            findable=True,
            alias=False,
            output=None,
            method_name="data",
            module="other.root.something.data",
            qualname="DataClass2",
        ),
    ]
    assert _categorize_functions() == {"something": {mocked_load.return_value}}



from __future__ import annotations

from datetime import timezone, tzinfo
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfoNotFoundError

from dissect.target.plugin import internal
from dissect.target.plugins.os.default.datetime import DateTimePlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


class UnixDateTimePlugin(DateTimePlugin):
    __namespace__ = "datetime"

    def __init__(self, target: Target):
        super().__init__(target)

        try:
            self._tzinfo = self.tz(target.timezone)
        except (TypeError, ZoneInfoNotFoundError):
            self.target.log.warning("Could not determine timezone of target, falling back to UTC for datetime helpers")
            self._tzinfo = timezone.utc

    def check_compatible(self) -> None:
        pass

    @internal(property=True)
    def tzinfo(self) -> tzinfo:
        """Return a ``datetime.tzinfo`` of the current system timezone."""
        return self._tzinfo



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

ModuleRecord = TargetRecordDescriptor(
    "linux/module",
    [
        ("string", "name"),
        ("varint", "size"),
        ("varint", "refcount"),
        ("string[]", "used_by"),
        ("path", "source"),
    ],
)


@dataclass
class Module:
    path: str
    name: str
    size: int
    refcnt: int
    used_by: list[str]


class ModulePlugin(Plugin):
    """Linux volatile kernel ``/sys/module`` plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._module_base_path = self.target.fs.path("/sys/module")

    def check_compatible(self) -> bool:
        if not self._module_base_path.is_dir() or not next(self._module_base_path.iterdir(), None):
            raise UnsupportedPluginError("No module paths found.")

    def _iterate_modules(self) -> Iterator[Module]:
        for module_path in self._module_base_path.iterdir():
            if module_path.joinpath("initstate").exists():
                holders = []
                if (holders_path := module_path.joinpath("holders")).exists():
                    holders = [item.name for item in holders_path.iterdir()]
                yield Module(
                    module_path,
                    module_path.name,
                    int(module_path.joinpath("coresize").read_text()),
                    int(module_path.joinpath("refcnt").read_text()),
                    holders,
                )

    @export(record=ModuleRecord)
    def sysmodules(self) -> Iterator[ModuleRecord]:
        """Return information about active kernel modules."""
        for module in self._iterate_modules():
            yield ModuleRecord(
                name=module.name,
                size=module.size,
                refcount=module.refcnt,
                used_by=module.used_by,
                source=module.path,
                _target=self.target,
            )

    @export(output="yield")
    def lsmod(self) -> Iterator[str]:
        """Return information about active kernel modules in lsmod format."""
        yield f"{'Module ':<28} {'Size':<7}  Used by"
        for module in self._iterate_modules():
            yield f"{module.name:<28} {module.size:<7}  {module.refcnt} {','.join(module.used_by)}"



from __future__ import annotations

import gzip
from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystems.qnxfs import QnxFilesystem, QnxFilesystemEntry
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.mark.parametrize(
    "filename",
    [
        "_data/filesystems/qnxfs/qnx4.bin.gz",
        "_data/filesystems/qnxfs/qnx6-be.bin.gz",
        "_data/filesystems/qnxfs/qnx6-le.bin.gz",
    ],
)
def test_qnxfs_detect(filename: str) -> None:
    """Test that QNX filesystems are correctly detected."""
    with gzip.open(absolute_path(filename), "rb") as fh:
        assert QnxFilesystem.detect(fh)
        assert QnxFilesystem(fh).qnxfs


@pytest.fixture
def qnx_fs() -> Iterator[QnxFilesystem]:
    with patch("dissect.qnxfs.QNXFS", return_value=Mock(block_size=4096)):
        qnx_fs = QnxFilesystem(Mock())
        yield qnx_fs


@pytest.fixture
def qnxfs_fs_file_entry(qnx_fs: QnxFilesystem) -> Iterator[QnxFilesystemEntry]:
    inode = Mock(
        mode=0o100664,
        inum=4,
        size=165002,
        uid=1000,
        gid=999,
        atime=datetime(2024, 10, 1, 12, 0, 0, tzinfo=timezone.utc),
        mtime=datetime(2024, 10, 2, 12, 0, 0, tzinfo=timezone.utc),
        ctime=datetime(2024, 10, 3, 12, 0, 0, tzinfo=timezone.utc),
        ftime=datetime(2024, 10, 4, 12, 0, 0, tzinfo=timezone.utc),
        nlink=2,
        is_file=lambda: True,
        is_dir=lambda: False,
        is_symlink=lambda: False,
    )

    return QnxFilesystemEntry(qnx_fs, "/some_file", inode)


@pytest.mark.parametrize(
    ("entry_fixture", "expected_blocks"),
    [("qnxfs_fs_file_entry", 323)],
)
def test_qnxfs_stat(entry_fixture: str, expected_blocks: int, request: pytest.FixtureRequest) -> None:
    """Test consistency in ``stat()`` results."""
    qnxfs_entry: QnxFilesystemEntry = request.getfixturevalue(entry_fixture)
    stat = qnxfs_entry.stat()

    entry = qnxfs_entry.entry
    assert stat.st_mode == entry.mode
    assert stat.st_ino == entry.inum
    assert stat.st_dev == id(qnxfs_entry.fs)
    assert stat.st_nlink == entry.nlink
    assert stat.st_uid == entry.uid
    assert stat.st_gid == entry.gid
    assert stat.st_size == entry.size
    assert stat.st_atime == entry.atime.timestamp()
    assert stat.st_mtime == entry.mtime.timestamp()
    assert stat.st_ctime == entry.ctime.timestamp()
    assert stat.st_blksize == qnxfs_entry.fs.qnxfs.block_size
    assert stat.st_blocks == expected_blocks



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.locate.plocate import PLocatePlugin, PLocateRecord
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_plocate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/var/lib/plocate/plocate.db", absolute_path("_data/plugins/os/unix/locate/plocate.db"))
    target_unix.add_plugin(PLocatePlugin)

    records = list(target_unix.plocate.locate())

    assert len(records) == 3594
    assert isinstance(records[0], type(PLocateRecord()))

    assert records[0].path.as_posix() == "/.dockerenv"
    assert records[1].path.as_posix() == "/bin"

    # regression for issue #696 (https://github.com/fox-it/dissect.target/issues/696)
    # records[31] would be `/etc/cron.daily/etc/debconf.conf` without the fix
    assert records[31].path.as_posix() == "/etc/cron.daily"
    assert records[32].path.as_posix() == "/etc/debconf.conf"



from __future__ import annotations

import stat
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.filesystem import FilesystemEntry, LayerFilesystemEntry
from dissect.target.helpers.magic import magic
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.filesystem.unix.capability import parse_entry as parse_capability_entry

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

FilesystemRecord = TargetRecordDescriptor(
    "filesystem/entry",
    [
        ("datetime", "atime"),
        ("datetime", "mtime"),
        ("datetime", "ctime"),
        ("datetime", "btime"),
        ("varint", "ino"),
        ("path", "path"),
        ("filesize", "size"),
        ("uint32", "mode"),
        ("uint32", "uid"),
        ("uint32", "gid"),
        ("string", "mimetype"),
        ("boolean", "is_suid"),
        ("string", "type"),
        ("string[]", "attr"),
        ("string[]", "fs_types"),
    ],
)


class WalkFsPlugin(Plugin):
    """Filesystem agnostic walkfs plugin."""

    def check_compatible(self) -> None:
        if not len(self.target.filesystems):
            raise UnsupportedPluginError("No filesystems to walk")

    @export(record=FilesystemRecord)
    @arg("--walkfs-path", default="/", help="path to recursively walk")
    @arg("--capability", action="store_true", help="output capability records")
    @arg("--mimetype", action="store_true", help="enable mimetype lookup of files")
    def walkfs(
        self, walkfs_path: str = "/", capability: bool = False, mimetype: bool = False
    ) -> Iterator[FilesystemRecord]:
        """Walk a target's filesystem and return all filesystem entries.

        References:
            - https://man7.org/linux/man-pages/man2/lstat.2.html
            - https://man7.org/linux/man-pages/man7/inode.7.html
            - https://man7.org/linux/man-pages/man7/xattr.7.html
            - https://man7.org/linux/man-pages/man2/execve.2.html
            - https://steflan-security.com/linux-privilege-escalation-suid-binaries
            - https://github.com/torvalds/linux/blob/master/include/uapi/linux/capability.h

        Yields FilesystemRecords for every filesystem entry and CapabilityRecords if ``xattr`` security
        attributes were found in the filesystem entry and the ``--capability`` flag is set.

        Mimetype lookup can be enabled using the ``--mimetype`` flag.

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            mtime (datetime): modified timestamp indicates the last time the contents of a file were modified.
            atime (datetime): access timestamp indicates the last time a file was accessed.
            ctime (datetime): changed timestamp indicates the last time metadata of a file was modified.
            btime (datetime): birth timestamp indicates the time when a file was created.
            ino (varint): number of the corresponding underlying filesystem inode.
            path (path): path location of the entry.
            size (filesize): size of the file in bytes on the filesystem.
            mode (uint32): contains the file type and mode.
            uid (uint32): the user id of the owner of the entry.
            gid (uint32): the group id of the owner of the entry.
            mimetype (string): detected mimetype of the entry.
            is_suid (boolean): denotes if the entry has the set-user-id bit set.
            attr (string[]): list of key-value pair attributes separated by '='.
            fs_types (string[]): list of filesystem type(s) of the entry.
        """

        path = self.target.fs.path(walkfs_path)

        if not path.exists():
            self.target.log.error("No such directory: '%s'", walkfs_path)
            return

        if not path.is_dir():
            self.target.log.error("Not a directory: '%s'", walkfs_path)
            return

        for entry in self.target.fs.recurse(walkfs_path):
            try:
                yield from generate_record(self.target, entry, capability, mimetype)
            except FileNotFoundError as e:  # noqa: PERF203
                self.target.log.warning("File not found: %s", entry)
                self.target.log.debug("", exc_info=e)
            except Exception as e:
                self.target.log.warning("Exception generating walkfs record for %s: %s", entry, e)
                self.target.log.debug("", exc_info=e)
                continue


def generate_record(
    target: Target, entry: FilesystemEntry, capability: bool, mimetype: bool
) -> Iterator[FilesystemRecord]:
    """Generate a :class:`WalkFsRecord` from the given :class:`FilesystemEntry`.

    Args:
        target: :class:`Target` instance
        entry: :class:`FilesystemEntry` instance
        capability: bool
        mimetype: bool

    Returns:
        Generator of :class:`FilesystemRecord` for the given :class:`FilesystemEntry`.
    """
    entry_stat = entry.lstat()

    if isinstance(entry, LayerFilesystemEntry):
        fs_types = [sub_entry.fs.__type__ for sub_entry in entry.entries]
    else:
        fs_types = [entry.fs.__type__]

    ftype = "unknown"
    if entry.is_symlink():
        ftype = "symlink"
    elif entry.is_dir(follow_symlinks=False):
        ftype = "dir"
    elif entry.is_file(follow_symlinks=False):
        ftype = "file"

    fields = {
        "atime": from_unix(entry_stat.st_atime),
        "mtime": from_unix(entry_stat.st_mtime),
        "ctime": from_unix(entry_stat.st_ctime),
        "btime": from_unix(entry_stat.st_birthtime) if entry_stat.st_birthtime else None,
        "ino": entry_stat.st_ino,
        "path": entry.path,
        "size": entry_stat.st_size,
        "mode": entry_stat.st_mode,
        "uid": entry_stat.st_uid,
        "gid": entry_stat.st_gid,
        "type": ftype,
        "is_suid": bool(entry_stat.st_mode & stat.S_ISUID),
        "fs_types": fs_types,
    }

    try:
        fields["attr"] = [f"{attr.name}={attr.value.hex()}" for attr in entry.lattr()]

    except (TypeError, AttributeError, NotImplementedError):
        # Suppress lattr calls on VirtualDirectory entries, filesystems without implemented attr's and NTFS attr's.
        pass

    except Exception as e:
        target.log.warning("Unable to expand xattr for entry %s: %s", entry.path, e)
        target.log.debug("", exc_info=e)

    if mimetype:
        try:
            fields["mimetype"] = magic.from_entry(entry, mime=True)
        except (FileNotFoundError, IsADirectoryError):
            pass
        except Exception as e:
            target.log.warning("Unable to determine mimetype for entry %s: %s", entry.path, e)
            target.log.debug("", exc_info=e)

    yield FilesystemRecord(
        **fields,
        _target=target,
    )

    if capability and fields.get("attr"):
        yield from parse_capability_entry(entry, target)



from __future__ import annotations

import zipfile
from collections import defaultdict
from typing import TYPE_CHECKING

from dissect.target.filesystem import LayerFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.filesystems.zip import ZipFilesystem
from dissect.target.helpers import loaderutil
from dissect.target.loader import Loader
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target

PREFIXES = ["", "fs"]


class DirLoader(Loader):
    """Load a directory as a filesystem."""

    @staticmethod
    def detect(path: Path) -> bool:
        return find_entry_path(path) is not None

    def map(self, target: Target) -> None:
        path = self.absolute_path.joinpath(find_entry_path(self.absolute_path))
        find_and_map_dirs(target, path)


def find_entry_path(path: Path) -> str | None:
    for prefix in PREFIXES:
        if find_dirs(path / prefix)[0] is not None:
            return prefix
    return None


def map_dirs(
    target: Target,
    dirs: list[Path | tuple[str, Path]],
    os_type: str,
    *,
    dirfs: type[DirectoryFilesystem] = DirectoryFilesystem,
    zipfs: type[ZipFilesystem] = ZipFilesystem,
    **kwargs,
) -> None:
    """Map directories as filesystems into the given target.

    Args:
        target: The target to map into.
        dirs: The directories to map as filesystems. If a list member is a tuple, the first element is the drive letter.
        os_type: The operating system type, used to determine how the filesystem should be mounted.
        dirfs: The filesystem class to use for directory filesystems.
        zipfs: The filesystem class to use for ZIP filesystems.
    """
    alt_separator = ""
    case_sensitive = True
    if os_type == OperatingSystem.WINDOWS:
        alt_separator = "\\"
        case_sensitive = False

    drive_letter_map = defaultdict(list)
    for path in dirs:
        drive_letter = None
        if isinstance(path, tuple):
            drive_letter, path = path
        elif is_drive_letter_path(path):
            drive_letter = path.name[0]

        if isinstance(path, zipfile.Path):
            dfs = zipfs(path.root.fp, path.at, alt_separator=alt_separator, case_sensitive=case_sensitive)
        else:
            dfs = dirfs(path, alt_separator=alt_separator, case_sensitive=case_sensitive)

        drive_letter_map[drive_letter].append(dfs)

    fs_to_add = []
    for drive_letter, dfs in drive_letter_map.items():
        if drive_letter is not None:
            if len(dfs) > 1:
                vfs = LayerFilesystem()
                for fs in dfs:
                    vfs.append_fs_layer(fs)
            else:
                vfs = dfs[0]

            fs_to_add.append(vfs)
            target.fs.mount(drive_letter.lower() + ":", vfs)
        else:
            fs_to_add.extend(dfs)

    for fs in fs_to_add:
        target.filesystems.add(fs)
        if os_type == OperatingSystem.WINDOWS:
            loaderutil.add_virtual_ntfs_filesystem(target, fs, **kwargs)


def find_and_map_dirs(target: Target, path: Path, **kwargs) -> None:
    """Try to find and map directories as filesystems into the given target.

    Args:
        target: The target to map into.
        path: The path to map from.
            If path is a local path, must be provided as an absolute path in order to work with ``target-mount``.
        **kwargs: Optional arguments for :func:`loaderutil.add_virtual_ntfs_filesystem
            <dissect.target.helpers.loaderutil.add_virtual_ntfs_filesystem>`.
    """
    os_type, dirs = find_dirs(path)

    map_dirs(target, dirs, os_type, **kwargs)


def find_dirs(path: Path) -> tuple[str, list[Path]]:
    """Try to find if ``path`` contains an operating system directory layout and return the OS type and detected
    directories.

    In the case of a Windows layout, try to find if there are directories for each drive letter and return
    them all.

    Args:
        path: The path to check.

    Returns:
        A tuple consisting of the found operating system layout and a list of all detected directories.
    """
    dirs = []
    os_type = None

    if path.is_dir():
        for p in path.iterdir():
            # Look for directories like C or C:
            if p.is_dir() and (is_drive_letter_path(p) or p.name in ("sysvol", "$rootfs$")):
                dirs.append(p)

                if not os_type:
                    os_type = os_type_from_path(p)

        if not os_type:
            os_type = os_type_from_path(path)
            dirs = [path]

    return os_type, dirs


def os_type_from_path(path: Path) -> OperatingSystem:
    """Try to detect what kind of operating system directory structure ``path`` contains.

    The operating system type is returned as a string.

    Args:
        path: The path to check.

    Returns:
        The detected operating system type.
    """
    if path.is_dir():
        dirlist = [p.name for p in path.iterdir()]
        dirlist_l = [e.lower() for e in dirlist]

        if "windows" in dirlist_l:
            windows_idx = dirlist_l.index("windows")
            windows_path = path.joinpath(dirlist[windows_idx])
            system32_exists = "system32" in [e.name.lower() for e in windows_path.iterdir()]
        else:
            system32_exists = False

        winnt_exists = "winnt" in dirlist_l

        if system32_exists or winnt_exists:  # Windows
            return OperatingSystem.WINDOWS

        etc_exists = "etc" in dirlist
        var_exists = "var" in dirlist
        library_exists = "Library" in dirlist

        if etc_exists and var_exists and not library_exists:  # Linux
            return OperatingSystem.LINUX

        if library_exists:  # OSX
            return OperatingSystem.OSX

    return None


def is_drive_letter_path(path: Path) -> bool:
    """Check if a path can be a drive letter, e.g. ``C`` or ``C:``.

    Args:
        path: The path to check.

    Returns:
        ``True`` if the path can be interpreted as a drive letter or ``False`` if it can't.
    """
    return len(path.name) == 1 or (len(path.name) == 2 and path.name[1] == ":")



from __future__ import annotations

from dissect.target.plugins.os.windows.dpapi.keyprovider.keyprovider import KeyProviderPlugin


class DefaultPasswordKeyProvider(KeyProviderPlugin):
    """DefaultPassword key provider plugin for DPAPI."""

    __namespace__ = "dpapi.keyprovider.defaultpassword"

    def check_compatible(self) -> None:
        return None



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystems.xfs import XfsFilesystem, XfsFilesystemEntry

if TYPE_CHECKING:
    from collections.abc import Iterator

NANOSECONDS_IN_SECOND = 1_000_000_000


@pytest.fixture
def xfs_fs() -> Iterator[XfsFilesystem]:
    with patch("dissect.xfs.xfs.XFS"):
        xfs_fs = XfsFilesystem(Mock())
        xfs_fs.xfs.block_size = 4096

        yield xfs_fs


@pytest.fixture
def xfs_fs_entry(xfs_fs: XfsFilesystem) -> XfsFilesystemEntry:
    atime = datetime(2024, 10, 1, 12, 0, 0, tzinfo=timezone.utc)
    mtime = datetime(2024, 10, 2, 12, 0, 0, tzinfo=timezone.utc)
    ctime = datetime(2024, 10, 3, 12, 0, 0, tzinfo=timezone.utc)
    crtime = datetime(2024, 10, 4, 12, 0, 0, tzinfo=timezone.utc)

    dinode = Mock(di_mode=0o100750, di_nlink=1, di_uid=1000, di_gid=999)
    inode = Mock(
        nblocks=10,
        inode=dinode,
        inum=4,
        size=4594,
        atime=atime,
        atime_ns=atime.timestamp() * NANOSECONDS_IN_SECOND,
        mtime=mtime,
        mtime_ns=mtime.timestamp() * NANOSECONDS_IN_SECOND,
        ctime=ctime,
        ctime_ns=ctime.timestamp() * NANOSECONDS_IN_SECOND,
        crtime=crtime,
        crtime_ns=crtime.timestamp() * NANOSECONDS_IN_SECOND,
    )
    return XfsFilesystemEntry(xfs_fs, "/some_file", inode)


def test_xfs_stat(xfs_fs: XfsFilesystem, xfs_fs_entry: XfsFilesystemEntry) -> None:
    stat = xfs_fs_entry.stat()

    entry = xfs_fs_entry.entry
    assert stat.st_mode == entry.inode.di_mode
    assert stat.st_ino == entry.inum
    assert stat.st_dev == id(xfs_fs)
    assert stat.st_nlink == entry.inode.di_nlink
    assert stat.st_uid == entry.inode.di_uid
    assert stat.st_gid == entry.inode.di_gid
    assert stat.st_size == entry.size
    assert stat.st_atime == entry.atime.timestamp()
    assert stat.st_atime_ns == entry.atime_ns
    assert stat.st_mtime == entry.mtime.timestamp()
    assert stat.st_mtime_ns == entry.mtime_ns
    assert stat.st_ctime == entry.ctime.timestamp()
    assert stat.st_ctime_ns == entry.ctime_ns
    assert stat.st_birthtime == entry.crtime.timestamp()
    assert stat.st_birthtime_ns == entry.crtime_ns
    assert stat.st_blksize == 4096
    assert stat.st_blocks == 10 * 4096 // 512



from __future__ import annotations

import io
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.target.exceptions import VolumeSystemError
from dissect.target.helpers import keychain
from dissect.target.helpers.lazy import import_lazy
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.utils import readinto

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.volumes.disk import DissectVolumeSystem

disk = import_lazy("dissect.target.volumes.disk")
"""A lazy import of :mod:`dissect.target.volumes.disk`."""
lvm = import_lazy("dissect.target.volumes.lvm")
"""A lazy import of :mod:`dissect.target.volumes.lvm`."""
vmfs = import_lazy("dissect.target.volumes.vmfs")
"""A lazy import of :mod:`dissect.target.volumes.vmfs`."""
md = import_lazy("dissect.target.volumes.md")
"""A lazy import of :mod:`dissect.target.volumes.md`."""
ddf = import_lazy("dissect.target.volumes.ddf")
"""A lazy import of :mod:`dissect.target.volumes.ddf`."""
bde = import_lazy("dissect.target.volumes.bde")
"""A lazy import of :mod:`dissect.target.volumes.bde`."""
luks = import_lazy("dissect.target.volumes.luks")
"""A lazy import of :mod:`dissect.target.volumes.luks`."""


log = get_logger(__name__)
"""A logger instance for this module."""

LOGICAL_VOLUME_MANAGERS: list[type[LogicalVolumeSystem]] = [
    lvm.LvmVolumeSystem,
    vmfs.VmfsVolumeSystem,
    md.MdVolumeSystem,
    ddf.DdfVolumeSystem,
]
"""All available :class:`LogicalVolumeSystem` classes."""
ENCRYPTED_VOLUME_MANAGERS: list[type[EncryptedVolumeSystem]] = [bde.BitlockerVolumeSystem, luks.LUKSVolumeSystem]
"""All available :class:`EncryptedVolumeSystem` classes."""


class VolumeSystem:
    """The base class for a volume system implementation.

    Volume systems are responsible for parsing a volume system over one or more disks and
    returning all available volumes.

    Subclasses of ``VolumeSystem`` must implement the ``_detect`` and ``_volumes`` methods.

    Args:
        fh: The source file-like object(s) on which to open the volume system.
        disk: A reference to the source disk or container. Defaults to ``fh`` if not provided.
        serial: Serial number of the volume system, if any.
    """

    # Due to lazy importing we generally can't use isinstance(), so we add a short identifying string to each class
    # This has the added benefit of having a readily available "pretty name" for each implementation
    __type__: str = None
    """A short string identifying the type of volume system."""

    def __init__(
        self, fh: BinaryIO | list[BinaryIO], disk: BinaryIO | list[BinaryIO] | None = None, serial: str | None = None
    ):
        self.fh = fh
        self.disk = disk or fh
        self.serial = serial
        self._volumes_list: list[Volume] = None

        if self.__type__ is None:
            raise NotImplementedError(f"{self.__class__.__name__} must define __type__")

    def __repr__(self) -> str:
        return f"<VolumeSystem type={self.__type__} serial={self.serial}>"

    @classmethod
    def detect(cls, fh: BinaryIO) -> bool:
        """Detects whether this ``VolumeSystem`` class can be opened on the given file-like object.

        The position of ``fh`` will be restored before returning.

        Returns:
            ``True`` or ``False`` if the ``VolumeSystem`` can be opened from this file-like object.
        """
        offset = fh.tell()
        try:
            fh.seek(0)
            return cls._detect(fh)
        except Exception as e:
            log.warning("Failed to detect %s volume system", cls)
            log.debug("", exc_info=e)
        finally:
            fh.seek(offset)

        return False

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detects whether this ``VolumeSystem`` class can be opened on the given file-like object.

        This method should be implemented by subclasses. The position of ``fh`` is guaranteed to be ``0``.

        Returns:
            ``True`` or ``False`` if the ``VolumeSystem`` can be opened from this file-like object.
        """
        raise NotImplementedError

    def _volumes(self) -> Iterator[Volume]:
        """List all valid discovered volumes found on the volume system.

        Returns:
            An iterator of all :class:`Volume` of the ``VolumeSystem``.
        """
        raise NotImplementedError

    @property
    def volumes(self) -> list[Volume]:
        """A list of all the discovered volumes."""
        if self._volumes_list is None:
            self._volumes_list = list(self._volumes())

        return self._volumes_list


class EncryptedVolumeSystem(VolumeSystem):
    """An extension of the :class:`VolumeSystem` class that provides additional functionality for
    dealing with encryption.

    It adds helper functions for interacting with the :attr:`~dissect.target.helpers.keychain.KEYCHAIN`,
    so that subclasses don't have to manually interact with it.

    Args:
        fh: The file-like object on which to open the encrypted volume system.
    """

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.keys = keychain.get_keys_for_provider(self.__type__) + keychain.get_keys_without_provider()

    def get_keys_for_identifier(self, identifier: str) -> list[keychain.Key]:
        """Retrieves a list of keys that match a single ``identifier``.

        Args:
            identifier: A single key identifier.

        Returns:
            All the keys for a single identifier.
        """
        return self.get_keys_for_identifiers([identifier])

    def get_keys_for_identifiers(self, identifiers: list[str]) -> list[keychain.Key]:
        """Retrieves a list of keys that match a list of ``identifiers``.

        Args:
            identifiers: A list of different key identifiers.
        """
        # normalise values before checks
        identifiers = [i.lower() for i in identifiers]
        return [key for key in self.keys if key.identifier and key.identifier.lower() in identifiers]

    def get_keys_without_identifier(self) -> list[keychain.Key]:
        """Retrieve a list of keys that have no identifier (``None``).

        These are the keys where no specific identifier was specified.
        """
        return [key for key in self.keys if key.identifier is None]


class LogicalVolumeSystem(VolumeSystem):
    """An extension of the :class:`VolumeSystem` class that provides additional functionality for dealing with
    logical volume systems.
    """

    @classmethod
    def detect_volume(cls, fh: BinaryIO) -> bool:
        """Determine whether the given file-like object belongs to this logical volume system.

        The position of ``fh`` will be restored before returning.

        Args:
            fh: A file-like object that may be part of the logical volume system.

        Returns:
            ``True`` if the given file-like object is part of the logical volume system, ``False`` otherwise.
        """
        offset = fh.tell()
        try:
            fh.seek(0)
            return cls._detect_volume(fh)
        except Exception as e:
            log.warning("Failed to detect %s logical volume", cls)
            log.debug("", exc_info=e)
        finally:
            fh.seek(offset)

        return False

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        """Determine whether the given file-like object belongs to this logical volume system.

        This method should be implemented by subclasses. The position of ``fh`` is guaranteed to be ``0``.

        Args:
            fh: A file-like object that may be part of the logical volume system.

        Returns:
            ``True`` if the given file-like object is part of the logical volume system, ``False`` otherwise.
        """
        raise NotImplementedError

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[Self]:
        """Open all the discovered logical volume systems from the given file-like objects.

        There can be more than one logical volume system on a given set of file-like objects. For example, you can have
        five disks or volumes with two separate LVM2 volume groups. This function is responsible for grouping
        the correct disks and volumes with each other, and correctly opening each distinct logical volume system.

        Args:
            volumes: A list of file-like objects to discover and open the logical volume systems on.

        Returns:
            An iterator of :class:`LogicalVolumeSystem`.
        """
        raise NotImplementedError


class Volume(io.IOBase):
    """A representation of a volume on disk.

    It behaves like a regular file-like object with some additional information bound to it.

    Args:
        fh: The raw file-like object of the volume.
        number: The logical volume number of this volume within the volume system.
        offset: Where the volume starts relative to the start of the volume system.
        size: The size of the volume.
        vtype: What kind of volume it is.
        name: The name of the volume.
        guid: The unique identifier of the volume.
        raw: A reference to the implementation specific object that the volume system uses for representing the volume.
        disk: A reference to the associated :class:`~dissect.target.container.Container`.
        vs: A reference to the associated :class:`VolumeSystem`.
        fs: A reference to the :class:`~dissect.target.filesystem.Filesystem` that is on this ``Volume``.
        drive_letter: The letter associated to the ``Volume``, such as `c` or `d` in Windows.
    """

    def __init__(
        self,
        fh: BinaryIO,
        number: int,
        offset: int | None,
        size: int,
        vtype: int | None,
        name: str | None,
        guid: str | None = None,
        raw: Any | None = None,
        disk: BinaryIO | list[BinaryIO] | None = None,
        vs: VolumeSystem | None = None,
        fs: Filesystem | None = None,
        drive_letter: str | None = None,
    ):
        self.fh = fh
        self.number = number
        self.offset = offset
        self.size = size
        self.type = vtype
        self.name = name
        self.guid = guid

        # Shorthand access to raw volume, disk, vs and fs objects
        self.raw = raw
        self.disk = disk
        self.vs = vs
        self.fs = fs
        self.drive_letter = drive_letter

        self.seek(0)

    def __repr__(self) -> str:
        return f"<Volume name={self.name!r} size={self.size!r} fs={self.fs.__type__ if self.fs else None!r}>"

    def read(self, length: int = -1) -> bytes:
        """Read a ``length`` of bytes from this ``Volume``."""
        return self.fh.read(length)

    def readinto(self, b: bytearray) -> int:
        """Uses :func:`dissect.target.helpers.utils.readinto`."""
        return readinto(buffer=b, fh=self)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """Change the stream position to ``offset``.

        ``whence`` determines where to seek from:

        * ``io.SEEK_SET`` (``0``):: absolute offset in the stream.
        * ``io.SEEK_CUR`` (``1``):: current position in the stream.
        * ``io.SEEK_END`` (``2``):: end of stream.

        Args:
            offset: The offset relative to the position indicated by ``whence``.
            whence: Where to start the seek from.
        """
        return self.fh.seek(offset, whence)

    def tell(self) -> int:
        """Returns the current seek position of the ``Volume``."""
        return self.fh.tell()

    def seekable(self) -> bool:
        """Returns whether ``seek`` can be used by this volume. Always ``True``."""
        return True


def open(fh: BinaryIO, *args, **kwargs) -> DissectVolumeSystem:
    """Open a :class:`~dissect.target.volumes.disk.DissectVolumeSystem` on the given file-like object.

    Args:
        fh: The file-like object to open a :class:`~dissect.target.volumes.disk.DissectVolumeSystem` on.

    Raises:
        VolumeSystemError: If opening the :class:`~dissect.target.volumes.disk.DissectVolumeSystem` failed.

    Returns:
        An opened :class:`~dissect.target.volumes.disk.DissectVolumeSystem`.
    """
    offset = fh.tell()
    fh.seek(0)

    try:
        return disk.DissectVolumeSystem(fh)
    except Exception as e:
        raise VolumeSystemError(f"Failed to load volume system for {fh}") from e
    finally:
        fh.seek(offset)


def is_lvm_volume(volume: BinaryIO) -> bool:
    """Determine whether the given file-like object belongs to any supported logical volume system.

    Args:
        volume: A file-like object to test if it is part of any logical volume system.
    """
    for logical_vs in LOGICAL_VOLUME_MANAGERS:
        try:
            if logical_vs.detect_volume(volume):
                return True
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", logical_vs)
            log.debug("", exc_info=e)
        except Exception as e:
            raise VolumeSystemError(f"Failed to detect logical volume for {volume}") from e

    return False


def is_encrypted(volume: BinaryIO) -> bool:
    """Determine whether the given file-like object belongs to any supported encrypted volume system.

    Args:
        volume: A file-like object to test if it is part of any encrypted volume system.
    """
    for manager in ENCRYPTED_VOLUME_MANAGERS:
        try:
            if manager.detect(volume):
                return True
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", manager)
            log.debug("", exc_info=e)
        except Exception as e:
            raise VolumeSystemError(f"Failed to detect encrypted volume for {volume}") from e
    return False


def open_encrypted(volume: BinaryIO) -> Iterator[Volume]:
    """Open an encrypted ``volume``.

    An encrypted volume can only be opened if the encrypted volume system can successfully decrypt the volume,
    meaning that the correct decryption key must be present in the :attr:`~dissect.target.helpers.keychain.KEYCHAIN`.

    The resulting :class:`Volume` object provides transparent decryption of the encrypted volume.

    Args:
        volume: A file-like object representing a :class:`Volume`.

    Returns:
        An iterator of decrypted :class:`Volume` objects as opened by the encrypted volume manager.
    """
    for manager_cls in ENCRYPTED_VOLUME_MANAGERS:
        try:
            if manager_cls.detect(volume):
                volume_manager = manager_cls(volume)
                yield from volume_manager.volumes
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", manager_cls)
            log.debug("", exc_info=e)
        except Exception as e:
            log.error(  # noqa: TRY400
                "Failed to open an encrypted volume %s with volume manager %s: %s", volume, manager_cls.__type__, e
            )
            log.debug("", exc_info=e)
    return None


def open_lvm(volumes: list[BinaryIO], *args, **kwargs) -> Iterator[VolumeSystem]:
    """Open a single logical volume system on a list of file-like objects.

    Args:
        volumes: A list of file-like objects to open a logical volume system on.

    Returns:
        An iterator of all the :class:`Volume` objects opened by the logical volume system.
    """
    for logical_vs in LOGICAL_VOLUME_MANAGERS:
        try:
            yield from logical_vs.open_all(volumes)
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", logical_vs)
            log.debug("", exc_info=e)
        except Exception as e:
            raise VolumeSystemError(f"Failed to load logical volume system for {volumes}") from e



from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree

from dissect.target import container
from dissect.target.helpers import fsutil
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class LibvirtLoader(Loader):
    """Load libvirt xml configuration files."""

    @staticmethod
    def detect(path: Path) -> bool:
        if path.suffix.lower() != ".xml":
            return False

        with path.open("rb") as fh:
            lines = fh.read(512).split(b"\n")
            # From what I've seen, these are are always at the start of the file
            # If its generated using virt-install
            needles = [b"<domain", b"<name>", b"<uuid>"]
            return all(any(needle in line for line in lines) for needle in needles)

    def map(self, target: Target) -> None:
        xml_data = ElementTree.fromstring(self.path.read_text())
        for disk in xml_data.findall("devices/disk/source"):
            if not (file := disk.get("file")):
                continue

            for part in [fsutil.basename(file), file]:
                if (path := self.base_path.joinpath(part)).exists():
                    target.disks.add(container.open(path))
                    break



from __future__ import annotations

from dissect.target.helpers.magic.magic import Magic, from_buffer, from_descriptor, from_entry, from_fh, from_file

__all__ = [
    "Magic",
    "from_buffer",
    "from_descriptor",
    "from_entry",
    "from_fh",
    "from_file",
]



from __future__ import annotations

import json as jsonlib

from dissect.target.helpers.docs import INDENT_STEP, get_docstring
from dissect.target.loader import LOADERS_BY_SCHEME
from dissect.target.plugin import Plugin, arg, export


class LoaderListPlugin(Plugin):
    """Plugin to list the available loaders."""

    def check_compatible(self) -> None:
        pass

    @export(output="none")
    @arg("-j", "--json", action="store_true", help="output in JSON format")
    def loaders(self, json: bool = False) -> None:
        """List the available loaders."""

        loaders_info = {}
        for key, loader in LOADERS_BY_SCHEME.items():
            try:
                docstring = get_docstring(loader, "No documentation.").splitlines()[0].strip()
                loaders_info[key] = docstring
            except ImportError:  # noqa: PERF203
                continue

        loaders = sorted(loaders_info.items())

        if json:
            print(jsonlib.dumps([{"name": name, "description": desc} for name, desc in loaders]), end="")

        else:
            print("Available loaders:")
            for loader_name, loader_description in loaders:
                print(f"{INDENT_STEP}{loader_name} - {loader_description}")



from __future__ import annotations

import gzip
import os
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.tools.yara import HAS_YARA
from dissect.target.tools.yara import main as target_yara
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
def test_yara(target_default: Target, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture) -> None:
    vfs = VirtualFilesystem()
    vfs.map_file_fh("test_file", BytesIO(b"hello there this is a test string!"))
    vfs.map_file_fh("/test/dir/to/test_file", BytesIO(b"this is another test string for YARA testing."))
    vfs.map_file_fh("should_not_hit", BytesIO(b"this is another file."))
    target_default.fs.mount("/", vfs)

    with patch("dissect.target.Target.open_all", return_value=[target_default]), monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-yara",
                "example.img",
                "--rules",
                str(absolute_path("_data/plugins/filesystem/yara/rule-dir/rule.yar")),
                "--path",
                "/",
                "--check",
                "-s",
            ],
        )
        target_yara()

        out, _ = capsys.readouterr()

        hit1 = "<filesystem/yara/match hostname=None domain=None ts_mtime=1970-01-01 00:00:00+00:00 path='/test_file' rule='test_rule_name' matches=['$=test string'] tags=['tag1', 'tag2', 'tag3'] digest=(md5=d690ba32b59d28614aebefe9b03c74d4, sha1=4b1ced217aabe37138e96fb93bf40026639b9d3b, sha256=7a644118588ff0dcf2fadbe198ae1f1629c29374bac491ba41d5cf957edf0dfc)"  # noqa E501
        hit2 = "<filesystem/yara/match hostname=None domain=None ts_mtime=1970-01-01 00:00:00+00:00 path='/test/dir/to/test_file' rule='test_rule_name' matches=['$=test string'] tags=['tag1', 'tag2', 'tag3'] digest=(md5=bd7490dd2978ce983e2e1613ac8444c0, sha1=849a062cf09280f5c7dce4c7f87c69a1d9262e08, sha256=9bf7629a67c7ce8019910f1c1251fe44b61b3fff55a59a5e148af3c207dc102f)"  # noqa E501

        assert len(out.splitlines()) == 2

        assert hit1 in out
        assert hit2 in out


@pytest.mark.skipif(not HAS_YARA, reason="requires yara-python")
@pytest.mark.parametrize("no_decompress", [False, True], ids=["decompress", "no-decompress"])
def test_yara_decompress(
    target_default: Target,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
    tmp_path: Path,
    no_decompress: bool,
) -> None:
    """Test that the yara plugin can automatically decompress files for scanning,
    and that this can be disabled with the --no-decompress flag."""
    vfs = VirtualFilesystem()
    target_default.fs.mount("/", vfs)

    gzip_path = tmp_path / "messages.gz"
    gzip_path.write_bytes(gzip.compress(b"test string for YARA testing (gzipped).", mtime=0))
    os.utime(gzip_path, (0, 0))  # ensure mtime is set to 0 for consistent testing
    vfs.map_file("/var/log/messages.gz", gzip_path)

    with (
        patch("dissect.target.Target.open_all", return_value=[target_default]),
        monkeypatch.context() as m,
    ):
        m.setattr(
            "sys.argv",
            [
                "target-yara",
                "example.img",
                "--rules",
                str(absolute_path("_data/plugins/filesystem/yara/rule-dir/rule.yar")),
                "--strings",
                "--no-decompress" if no_decompress else "",
            ],
        )
        target_yara()

        out, _ = capsys.readouterr()

        hit = "<filesystem/yara/match hostname=None domain=None ts_mtime=1970-01-01 00:00:00+00:00 path='/var/log/messages.gz' rule='test_rule_name' matches=['$=test string'] tags=['tag1', 'tag2', 'tag3'] digest=(md5=485a4c5e37cd08a0cdf028cc2b1f32b4, sha1=7716bff4d3282184a17f35f5f0dde25aede762f9, sha256=d175bb08145c0be3338b058f8b7a775cc08967ba6aa7448fa2af9957eb1ab37f)"  # noqa E501

        if no_decompress:
            assert len(out.splitlines()) == 0
            assert hit not in out
        else:
            assert len(out.splitlines()) == 1
            assert hit in out



from __future__ import annotations

import re
from datetime import datetime, timedelta, timezone
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugins.apps.webserver.apache import ApachePlugin
from dissect.target.plugins.apps.webserver.citrix import (
    LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME,
    LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME_WITH_HEADERS,
    CitrixWebserverPlugin,
)
from dissect.target.plugins.apps.webserver.webserver import WebserverPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_infer_access_log_citrix_netscaler_combined_response_time() -> None:
    log_combined = (
        '127.0.0.1 - - [11/Aug/2023:08:30:00 +0000] [82775] "GET / HTTP/1.1" 200 18705 "-" "curl/7.78.0" "Time: 39770 '
        'microsecs"'
    )
    logformat = CitrixWebserverPlugin.infer_access_log_format(log_combined)
    assert logformat == LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME


def test_infer_access_log_citrix_netscaler_combined_response_time_with_headers() -> None:
    log_combined_with_headers = (
        '1.2.3.4 -> 5.6.7.8 - - [19/Dec/2022:17:25:48 +0100] [69420] "GET / HTTP/1.1" 200 1436 "-" "Mozilla/5.0 '
        '(Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36" '
        '"Time: 58774 microsecs"'
    )
    logformat = CitrixWebserverPlugin.infer_access_log_format(log_combined_with_headers)
    assert logformat == LOG_FORMAT_CITRIX_NETSCALER_ACCESS_COMBINED_RESPONSE_TIME_WITH_HEADERS


def test_access_logs(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    tz = timezone(timedelta(hours=1))
    data_file = absolute_path("_data/plugins/apps/webserver/citrix/httpaccess.log")
    fs_bsd.map_file("var/log/httpaccess.log", data_file)

    target_citrix.add_plugin(CitrixWebserverPlugin)

    results = list(target_citrix.citrix.access())

    citrix_netscaler_headers_combined_response_log = results[0]
    assert citrix_netscaler_headers_combined_response_log.ts == datetime(2022, 12, 19, 17, 25, 48, tzinfo=tz)
    assert citrix_netscaler_headers_combined_response_log.remote_ip == "1.2.3.4"
    assert citrix_netscaler_headers_combined_response_log.local_ip == "5.6.7.8"
    assert citrix_netscaler_headers_combined_response_log.status_code == 200
    assert citrix_netscaler_headers_combined_response_log.method == "GET"
    assert citrix_netscaler_headers_combined_response_log.uri == "/"
    assert citrix_netscaler_headers_combined_response_log.protocol == "HTTP/1.1"
    assert citrix_netscaler_headers_combined_response_log.referer is None
    assert (
        citrix_netscaler_headers_combined_response_log.useragent
        == "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 "
        "Safari/537.36"
    )
    assert citrix_netscaler_headers_combined_response_log.response_time_ms == 58
    assert citrix_netscaler_headers_combined_response_log.pid == 69420

    citrix_netscaler_headers_combined_response_log = results[1]
    assert citrix_netscaler_headers_combined_response_log.ts == datetime(2023, 8, 11, 8, 30, 0, tzinfo=timezone.utc)
    assert citrix_netscaler_headers_combined_response_log.remote_ip == "127.0.0.1"
    assert citrix_netscaler_headers_combined_response_log.status_code == 200
    assert citrix_netscaler_headers_combined_response_log.method == "GET"
    assert citrix_netscaler_headers_combined_response_log.uri == "/"
    assert citrix_netscaler_headers_combined_response_log.protocol == "HTTP/1.1"
    assert citrix_netscaler_headers_combined_response_log.referer is None
    assert citrix_netscaler_headers_combined_response_log.useragent == "curl/7.78.0"
    assert citrix_netscaler_headers_combined_response_log.response_time_ms == 39
    assert citrix_netscaler_headers_combined_response_log.pid == 82775


def test_error_logs(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    fs_bsd.map_file("var/log/httperror-vpn.log", BytesIO(b"Foo"))
    fs_bsd.map_file("var/log/httperror.log", BytesIO(b"Bar"))

    target_citrix.add_plugin(CitrixWebserverPlugin)

    assert len(target_citrix.citrix.error_paths) == 2


def test_access_logs_webserver_namespace(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/citrix/httpaccess.log")
    fs_bsd.map_file("var/log/httpaccess.log", data_file)

    with pytest.raises(
        UnsupportedPluginError, match=re.escape("Use the 'apps.webserver.citrix' apache plugin instead")
    ):
        target_citrix.add_plugin(ApachePlugin)

    target_citrix.add_plugin(CitrixWebserverPlugin)
    target_citrix.add_plugin(WebserverPlugin)

    results_via_webserver_namespace = list(target_citrix.webserver.access())
    results_via_citrix_namespace = list(target_citrix.citrix.access())

    assert len(results_via_webserver_namespace) == 2
    assert len(results_via_citrix_namespace) == 2
    assert [str(r) for r in results_via_webserver_namespace] == [str(r) for r in results_via_citrix_namespace]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

RunKeyRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/registry/run",
    [
        ("datetime", "ts"),
        ("wstring", "name"),
        ("command", "command"),
        ("string", "key"),
    ],
)


class RunKeysPlugin(Plugin):
    """Plugin that iterates various Runkey locations."""

    KEYS = (
        "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer\\Run",
        "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnce",
        "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnce\\Setup",
        "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx",
        "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnce",
        "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnce\\Setup",
        "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx",
        "HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer\\Run",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer\\Run",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnce",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnce\\Setup",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx",
        "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer\\Run",
        "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnce",
        "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnce\\Setup",
        "HKEY_CURRENT_USER\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\RunOnceEx",
    )

    def check_compatible(self) -> None:
        if not next(self.target.registry.keys(self.KEYS), None):
            raise UnsupportedPluginError("No registry run key found")

    @export(record=RunKeyRecord)
    def runkeys(self) -> Iterator[RunKeyRecord]:
        """Iterate various run key locations. See source for all locations.

        Run keys (Run and RunOnce) are registry keys that make a program run when a user logs on. a Run key runs every
        time the user logs on and the RunOnce key makes the program run once and deletes the key after. Often leveraged
        as a persistence mechanism.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/setupapi/run-and-runonce-registry-keys

        Yields RunKeyRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The registry key last modified timestamp.
            name (string): The run key name.
            command (command): The run key command.
            key (string): The source key for this run key.
        """
        for key in self.KEYS:
            for r in self.target.registry.keys(key):
                user = self.target.registry.get_user(r)
                for entry in r.values():
                    yield RunKeyRecord(
                        ts=r.ts,
                        name=entry.name,
                        command=entry.value or None,
                        key=key,
                        _target=self.target,
                        _key=r,
                        _user=user,
                    )



from __future__ import annotations

import logging
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.helpers.logging import TRACE_LEVEL
from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.registry import RegistryPlugin
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterable

    from dissect.target.filesystem import VirtualFilesystem


def test_missing_hives(fs_win: VirtualFilesystem, caplog: pytest.LogCaptureFixture) -> None:
    target = Target()
    target.filesystems.add(fs_win)

    with caplog.at_level(TRACE_LEVEL, target.log.name):
        target.apply()

        expected = []

        base_paths = [
            "sysvol/windows/system32/config",
            "sysvol/WINNT/system32/config",
            "sysvol/windows",
            "sysvol/reactos",
            "sysvol/windows/system32/config/RegBack",
        ]

        for base_path in base_paths:
            expected += [f"{target}: Could not find hive: {base_path}/{hive}" for hive in RegistryPlugin.SYSTEM]

        assert [record.message for record in caplog.records if record.filename == "registry.py"] == expected


def test_missing_user_hives(
    fs_win: VirtualFilesystem, target_win_users: Target, caplog: pytest.LogCaptureFixture
) -> None:
    fs_win.makedirs("Users/John")

    with caplog.at_level(logging.DEBUG, target_win_users.log.name):
        target_win_users.registry.load_user_hives()

        assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
            f"{target_win_users}: Could not find ntuser.dat: C:/Users/John/ntuser.dat",
            f"{target_win_users}: Could not find usrclass.dat: C:/Users/John/AppData/Local/Microsoft/Windows/usrclass.dat",  # noqa: E501
        ]


def test_empty_hives(fs_win: VirtualFilesystem, caplog: pytest.LogCaptureFixture) -> None:
    fs_win.map_file_fh("windows/system32/config/SYSTEM", BytesIO())
    fs_win.map_file_fh("boot/BCD", BytesIO())

    target = Target()
    target.filesystems.add(fs_win)

    with caplog.at_level(logging.WARNING, target.log.name):
        target.apply()

        assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
            f"{target}: Empty hive: sysvol/windows/system32/config/SYSTEM",
            f"{target}: Empty BCD hive: sysvol/boot/BCD",
        ]


def test_empty_hives_skip_warning(fs_win: VirtualFilesystem, caplog: pytest.LogCaptureFixture) -> None:
    fake_fh = BytesIO()
    fake_fh.size = 1
    fs_win.map_file_fh("windows/system32/config/SYSTEM", fake_fh)
    fs_win.map_file_fh("windows/system32/config/RegBack/SYSTEM", BytesIO())

    target = Target()
    target.filesystems.add(fs_win)

    with (
        caplog.at_level(logging.WARNING, target.log.name),
        patch("dissect.target.plugins.os.windows.registry.RegfHive"),
    ):
        target.apply()

        assert [record.message for record in caplog.records if record.filename == "registry.py"] == []


def test_empty_user_hives(
    fs_win: VirtualFilesystem, target_win_users: Target, caplog: pytest.LogCaptureFixture
) -> None:
    fs_win.map_file_fh("Users/John/ntuser.dat", BytesIO())
    fs_win.map_file_fh("Users/John/AppData/Local/Microsoft/Windows/usrclass.dat", BytesIO())

    with caplog.at_level(logging.WARNING, target_win_users.log.name):
        target_win_users.registry.load_user_hives()

        assert [record.message for record in caplog.records if record.filename == "registry.py"] == [
            f"{target_win_users}: Empty NTUSER.DAT hive: C:/Users/John/ntuser.dat",
            f"{target_win_users}: Empty UsrClass.DAT hive: C:/Users/John/AppData/Local/Microsoft/Windows/usrclass.dat",
        ]


@pytest.mark.parametrize(
    ("pattern", "key_names"),
    [
        (
            "\\HKLM\\SOFTWARE\\Microsoft",
            ["Microsoft"],
        ),
        (
            "\\*\\SOFTWARE\\Microsoft",
            ["Microsoft"],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList\\*\\",
            [
                "S-1-5-18",
                "S-1-5-21-3263113198-3007035898-945866154-1002",
            ],
        ),
        (
            "\\HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows XP\\*",
            [],
        ),
        (
            "\\*",
            [
                "HKEY_LOCAL_MACHINE",
                "HKEY_USERS",
            ],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\*\\CurrentVersion\\ProfileList",
            ["ProfileList"],
        ),
        (
            "HKEY_LOCAL_MACHINE\\SOFTWARE\\*\\Windows NT\\*\\ProfileList\\",
            ["ProfileList"],
        ),
    ],
)
def test_registry_plugin_glob_ext(target_win_users: Target, pattern: str, key_names: list[str]) -> None:
    registry_plugin = target_win_users.registry

    key_collections = registry_plugin.glob_ext(pattern)
    collection_names = [key_collection.name for key_collection in key_collections]

    assert sorted(collection_names) == sorted(key_names)


def test_registry_plugin_root_none(target_win_users: Target) -> None:
    plugin: RegistryPlugin = target_win_users.registry

    assert plugin.key()
    assert plugin.key("")


@pytest.mark.parametrize(
    ("keys", "values", "expected_output"),
    [
        pytest.param(
            "HKLM\\SOFTWARE\\SomePath",
            "Foo",
            ["FooValue"],
            id="single-key-single-value",
        ),
        pytest.param(
            "HKLM\\SOFTWARE\\SomePath",
            ("Foo", "Bar"),
            ["FooValue", "BarValue"],
            id="single-key-multi-value",
        ),
        pytest.param(
            ("HKLM\\SOFTWARE\\SomePath", "HKLM\\SOFTWARE\\AnotherPath"),
            "Foo",
            ["FooValue", "AnotherFooValue"],
            id="multi-key-single-value",
        ),
        pytest.param(
            ("HKLM\\SOFTWARE\\SomePath", "HKLM\\SOFTWARE\\AnotherPath"),
            ("Foo", "Bar"),
            ["FooValue", "BarValue", "AnotherFooValue", "AnotherBarValue"],
            id="multi-key-multi-value",
        ),
    ],
)
def test_registry_plugin_values_keys(
    target_win_users: Target,
    hive_hklm: VirtualHive,
    keys: str | Iterable[str] | None,
    values: str | Iterable[str] | None,
    expected_output: list[str],
) -> None:
    """Test if we can handle different input values for :meth:`RegistryPlugin.values`."""

    key_path = "SOFTWARE\\SomePath"
    key = VirtualKey(hive_hklm, key_path)
    key.add_value("Foo", VirtualValue(hive_hklm, "Foo", "FooValue"))
    key.add_value("Bar", VirtualValue(hive_hklm, "Bar", "BarValue"))
    hive_hklm.map_key(key_path, key)

    key_path = "SOFTWARE\\AnotherPath"
    key = VirtualKey(hive_hklm, key_path)
    key.add_value("Foo", VirtualValue(hive_hklm, "Foo", "AnotherFooValue"))
    key.add_value("Bar", VirtualValue(hive_hklm, "Bar", "AnotherBarValue"))
    hive_hklm.map_key(key_path, key)

    assert [v.value for v in target_win_users.registry.values(keys, values)] == expected_output



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

bam_def = """
    struct entry {
        uint64 ts;
    };
    """
c_bam = cstruct().load(bam_def)

BamDamRecord = TargetRecordDescriptor(
    "windows/registry/bam",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)


class BamDamPlugin(Plugin):
    """Plugin for bam/dam registry keys."""

    KEYS = (
        "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\bam\\UserSettings",
        "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\dam\\UserSettings",
        "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\bam\\State\\UserSettings",
        "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\dam\\State\\UserSettings",
    )

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEYS))) > 0:
            raise UnsupportedPluginError("No bam or dam registry keys not found")

    @export(record=BamDamRecord)
    def bam(self) -> Iterator[BamDamRecord]:
        """Parse bam and dam registry keys.

        Yields BamDamRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The parsed timestamp.
            path (uri): The parsed path.
        """
        for reg_path in self.KEYS:
            for r in self.target.registry.keys(reg_path):
                for sub in r.subkeys():
                    for entry in sub.values():
                        if isinstance(entry.value, int):
                            continue

                        data = c_bam.entry(entry.value)
                        yield BamDamRecord(
                            ts=wintimestamp(data.ts),
                            path=self.target.fs.path(entry.name),
                            _target=self.target,
                        )



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.volume.ddf.ddf import DDF, DEFAULT_SECTOR_SIZE, DDFPhysicalDisk

from dissect.target.volume import LogicalVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self


class DdfVolumeSystem(LogicalVolumeSystem):
    __type__ = "ddf"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        self.ddf = DDF(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[Self]:
        sets: dict[bytes, list[DDFPhysicalDisk]] = {}

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            disk = DDFPhysicalDisk(vol)
            sets.setdefault(disk.anchor.DDF_Header_GUID, []).append(disk)

        for devs in sets.values():
            try:
                yield cls(devs)
            except Exception:  # noqa: PERF203
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        vols = [fh] if not isinstance(fh, list) else fh
        return any(DdfVolumeSystem.detect_volume(vol) for vol in vols)

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        fh.seek(-DEFAULT_SECTOR_SIZE, io.SEEK_END)
        return int.from_bytes(fh.read(4), "big") == 0xDE11DE11

    def _volumes(self) -> Iterator[Volume]:
        # MD only supports one configuration and virtual disk but doing this as a loop
        # makes it automatically safe for empty configurations
        for conf in self.ddf.configurations:
            for vd in conf.virtual_disks:
                fh = vd.open()
                yield Volume(fh, 1, None, vd.size, None, vd.name, vd.uuid, raw=self.ddf, disk=self.disk, vs=self)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux._os import LinuxPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class RedHatPlugin(LinuxPlugin):
    """RedHat, CentOS and Fedora Plugin."""

    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        REDHAT_PATHS = {
            "/etc/centos-release",
            "/etc/fedora-release",
            "/etc/redhat-release",
            "/etc/sysconfig/network-scripts",  # legacy detection
        }

        for fs in target.filesystems:
            for path in REDHAT_PATHS:
                if fs.exists(path):
                    return fs
        return None
